import os

from six import iteritems

from docplex.mp.format import ExchangeFormat
from docplex.mp.utils import is_string


# some utility methods
def _get_value_as_int(d, option):
    try:
        value = int(d[option])
    except Exception:
        value = None
    return value

def _get_value_as_string(d, option):
    return d.get(option, None)

def _get_value_as_boolean(d, option):
    try:
        value = _convert_to_bool(d[option])
    except Exception:
        value = None
    return value

_BOOLEAN_STATES = {'1': True, 'yes': True, 'true': True, 'on': True,
                   '0': False, 'no': False, 'false': False, 'off': False}

def _convert_to_bool(value):
    if value.lower() not in _BOOLEAN_STATES:
        raise ValueError('Not a boolean: %s' % value)
    return _BOOLEAN_STATES[value.lower()]



class DOcloudContext(object):
    """ The context to solve models on DOcloud.
    
    The context contains properties that alter the behavior of the DOcloud
    connector.
    
    Attributes:
        url: The DOcloud service URL.
        api_key: The DOcloud service API key to use.
        display_logs_after_run: If True, logs are downloaded and printed after
            the model is run.
        run_deterministic: Specific engine parameters are uploaded to keep the run deterministic.
        verbose: Makes the connector verbose.
        timeout: The timeout for requests.
        waittime: The wait time to wait for jobs to finish.
        verify: If True, verifies SSL certificates.
        log_requests: When True, the REST requests are logged.
        exchange_format: The exchange format to use.
           When setting the format, you can use the following strings: "lp".
           When getting the format, the property type is docplex.mp.format.ExchangeFormat.
    """
    def __init__(self, url=None, api_key=None):
        """ Creates a new DOcloud context.
        """
        self.url = url
        self.api_key = api_key
        self.display_logs_after_run = False
        self.run_deterministic = False
        self.verbose = False
        self.timeout = None
        self.waittime = None
        self.verify = None  # default is None so that we use defaults
        self.log_requests = None
        self._exchange_format = None
        self.debug_dump = None
        self.debug_dump_dir = None

    def copy(self):
        from copy import copy as shallow_copy
        return shallow_copy(self)




    def get_exchange_format(self):
        return self._exchange_format
    def set_exchange_format(self, exchange_format):
        if exchange_format is not None:
            self._exchange_format = ExchangeFormat.fromstring(exchange_format)
        else:
            self._exchange_format = None
    exchange_format = property(get_exchange_format, set_exchange_format)

    def has_credentials(self):
        """Returns True if this context has valid credentials.
        """
        if not self.url:
            return False
        elif not is_string(self.url):
            print("DOcloud: URL is not a string: {0!s}".format(self.url))
            return False
        elif not self.url.startswith("http"):
            print("DOcloud: not a valid URL: {0!s}, expecting https://...".format(self.url))
            return False
        if not self.api_key:
            return False
        elif not is_string(self.api_key):
            print("API key is not a string: {0!s}".format(self.api_key))
            return False

        return True


    def print_information(self):
        print(self.to_string())

    def to_string(self):
        quoted_url = "\"{}\"".format(self.url) if isinstance(self.url, str) else str(self.url)
        quoted_key = "\"{}\"".format(self.api_key[:4]+"*******"+self.api_key[len(self.api_key)-4:]) if isinstance(self.api_key, str) else str(self.api_key)
        t_out = str(self.timeout)
        return "context<url={0}, auth={1}, timeout={2}>".format(quoted_url, quoted_key, t_out)

    def __str__(self):
        return self.to_string()
    
    def read_from_rcfile(self, filename):
        """Reads this context from the given ressource file.
        
        The specified ressource file contains `name=value` pairs. For example::
        
           url = "https://docloud.service.com/job_manager/rest/v1"
           api_key = "example api_key"
           
        """
        list_of_properties = []
        with open(filename, "r") as f:
            current_name = None
            current_value = None
            lineno = 0
            for l in f:
                lineno += 1
                l = l.strip()
                if len(l) == 0:
                    continue
                if l[-1] is '\\':
                    s = l[:-1]
                else:
                    s = l
                # process case of continuation
                if current_value is not None:
                    current_value += s
                else:
                    if s.startswith("#"):
                        continue
                    spl = s.split("=", 1)
                    if len(spl) != 2:
                        # try with ":"
                        spl = s.split(":", 1)
                        if len(spl) != 2:
                            raise ValueError("Syntax error in line %s" % lineno)
                    current_name = spl[0].strip()
                    current_value = spl[1].strip()
                # continuation ?
                if l[-1] is not '\\':
                    list_of_properties.append((current_name, current_value))
                    current_name = None
                    current_value = None
            # cases where last line of file ends with '\'
            if current_name is not None:
                list_of_properties.append((current_name, current_value))
        self.update_from_list(list_of_properties)

    def update_from_python_module(self, py_module):
        for k, v in iteritems(py_module.__dict__):
            if k.startswith("__"):
                pass
            elif hasattr(self, k):
                setattr(self, k, v)
        return self
    
    def update_from_list(self, values):
        """Updates this context from the list of properties specified as 
        (name, value) tuples.
        """
        d = {}
        for name, v in values:
            d[name] = v
       
        url = _get_value_as_string(d, "url")
        if url is not None:
            self.url = _get_value_as_string(d, "url")
       
        api_key = _get_value_as_string(d, "api_key")
        if api_key is not None:
            self.api_key = api_key
            
        display_logs_after_run = _get_value_as_boolean(d, "display_logs_after_run")
        
        if display_logs_after_run is not None:
            self.display_logs_after_run = display_logs_after_run
            
            
        run_deterministic = _get_value_as_boolean(d, "run_deterministric")
        if run_deterministic is not None:
            self.run_deterministic = run_deterministic
        
        verbose = _get_value_as_boolean(d, "verbose")
        if verbose is not None:
            self.verbose = verbose
        
        timeout = _get_value_as_int(d, "timeout")
        if timeout is not None:
            self.timeout = timeout
            
        waittime = _get_value_as_int(d, "waittime")
        if waittime is not None:
            self.waittime = waittime

        debug_dump = _get_value_as_boolean(d, "debug_dump")
        if debug_dump is not None:
            self.debug_dump = debug_dump
        
        debug_dump_dir = _get_value_as_string(d, "debug_dump_dir")
        if debug_dump_dir is not None:
            self.debug_dump_dir = debug_dump_dir

        verify = _get_value_as_boolean(d, "verify")
        # defaults to True, and we need to test against None to differentiate from False
        self.verify = verify if verify is not None else True

        log_requests = _get_value_as_boolean(d, "log_requests")
        if log_requests is not None:
            self.log_requests = log_requests

        xformat = _get_value_as_string(d, "exchange_format")
        if xformat is not None:
            self.exchange_format = xformat  # None or other stuff handled in property


    @staticmethod
    def make_from_config_parser(config, section_name=None):
        """Reads this context from the given ConfigParser.
        
        Config is read from section 'DOcloudConnector'.
        
        Recognized options have the same name as the attributes of ``DOcloudContext``.
        """
        try:
            import configparser
        except ImportError:
            try:
                import ConfigParser
            except Exception as e:
                raise Exception("readFromConfigParser works only when configparser is installed")
        
        context = DOcloudContext()
        context.update_from_list(config.items(section_name))
        return context

    @staticmethod
    def make_default_context(url, api_key=None):
        context = DOcloudContext(url, api_key)
        if api_key is None:
            home = os.path.expanduser("~")
            rcfile = os.path.join(home, ".docplexrc")
            if os.path.isfile(rcfile):
                context.read_from_rcfile(rcfile)
        return context

    @staticmethod
    def make_context_from_python(pyconfig_module_path):
        """Builds this context from a Python initialization file.

        Recognized options have the same name as the attributes of ``DOcloudContext``.
        """
        try:
            from importlib import import_module

            config_module = import_module(pyconfig_module_path)
        except ImportError:
            raise Exception("make_context_from_python: cannot import: {0}".format(pyconfig_module_path))
        #
        context = DOcloudContext()
        return context.update_from_python_module(config_module)


 