import sys

from docplex.mp.engine import DummyEngine
from docplex.mp.utils import is_number, is_iterable, generate_constant
import docplex.mp.linear as pyo_ln
from docplex.mp.linear import VarType, AbstractLinearConstraint, IndicatorConstraint
from docplex.mp.progress import ProgressData
from docplex.mp.solution import SolveSolution
import cplex

from six import iteritems
from contextlib import contextmanager

from enum import Enum
# CHECK THIS
# noinspection PyProtectedMember
from cplex._internal import _subinterfaces
from cplex.callbacks import MIPInfoCallback
from math import fabs

# noinspection PyProtectedMember
import cplex._internal._constants as cpx_cst
from cplex.exceptions import CplexError

from docplex.mp.utils import fast_range

# gendoc: ignore


class ConnectListenersCallback(MIPInfoCallback):
    """ This class is piecewise.
    """

    RELATIVE_EPS = 1e-5
    ABS_EPS = 1e-4

    # noinspection PyAttributeOutsideInit
    def initialize(self, listeners, node_period=-1):
        self.__listeners = listeners
        self.__pdata = ProgressData()
        self.__last_printed_node = 0
        self._node_period = node_period
        self._start_time = -1
        self._start_dettime = -1


    @property
    def last_incumbent(self):
        return self.__pdata.current_objective

    @property
    def last_best_bound(self):
        return self.__pdata.best_bound

    @property
    def last_nb_nodes(self):
        return self.__pdata.current_nb_nodes

    def _has_changed(self, old_value, new_value):
        """ Internal logic to filter non-significant changes."""
        abs_diff = fabs(new_value - old_value)
        rel_diff = abs_diff / (1.0 + fabs(new_value))
        return rel_diff >= self.RELATIVE_EPS or abs_diff >= self.ABS_EPS

    def __call__(self):
        do_call = 0
        has_incumbent = self.has_incumbent()
        old_incumbent_obj = self.last_incumbent
        new_incumbent_obj = old_incumbent_obj
        if has_incumbent:
            new_incumbent_obj = self.get_incumbent_objective_value()
            if self._has_changed(old_incumbent_obj, new_incumbent_obj):
                do_call += 1
        # see with bound change ??

        nb_nodes_so_far = self.get_num_nodes()
        current_best_bound = self.get_best_objective_value()

        if self._node_period > 1 and nb_nodes_so_far >= self.__last_printed_node + self._node_period:
            do_call += 4

        if self._start_time < 0:
            self._start_time = self.get_start_time()
        if self._start_dettime < 0:
            self._start_dettime = self.get_start_dettime()

        pdata = self.__pdata
        pdata.is_incumbent = has_incumbent
        pdata.current_objective = new_incumbent_obj
        pdata.best_bound = current_best_bound
        pdata.mip_gap = self.get_MIP_relative_gap()
        pdata.current_nb_nodes = nb_nodes_so_far
        pdata.remaining_nb_nodes = self.get_num_remaining_nodes()
        pdata.time = self.get_time() - self._start_time
        pdata.det_time = self.get_dettime() - self._start_dettime
        if do_call:
            self.__last_printed_node = nb_nodes_so_far
            for l in self.__listeners:
                l.notify_progress(pdata)


class CplexIndexMode(Enum):
    # enumerated value for different ways to handle cplex indices.
    # Query is old-way: ask Cplex by name, very slow.
    #   Guess means compute guessed indices, do not query
    #   UseReturn is new API: assume "add" returns a range and use it.
    Query, Guess, UseReturn = 1, 2, 3

    _mode2string = {"quey": Query, "guess": Guess, "return": UseReturn}

    @staticmethod
    def parse(text, default_mode):
        if not text:
            return default_mode
        else:
            return CplexIndexMode._mode2string.get(text.lower(), default_mode)


# noinspection PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember,PyProtectedMember
class CplexEngine(DummyEngine):
    """
        Cplex engine wrapper.
    """

    def __init__(self, error_handler, index_mode=None, docloud_context=None):
        """
        INTERNAL
        :param error_handler: an error handler instance, typically the model's
        :param index_mode: a string describing how cplex indices are to be managed
        :param docloud_context: why is this here????
        :return:
        """
        DummyEngine.__init__(self)
        cpx = cplex.Cplex()
        cpx.set_problem_name('Model')
        self.__errorHandler = error_handler
        self.__print_cplex_msgs = True  # initial state
        self._index_mode = CplexIndexMode.parse(index_mode, CplexIndexMode.Guess)

        # cpx.parameters.read.datacheck.set(1)
        self.__cplex = cpx

        self.__solveCount = 0
        self.__lastSolveStatus = False
        self.__lastObjectiveValue = -cplex.infinity

        # remember truly allocated indices
        self._last_used_ct_index = -1
        self._last_used_var_index = -1

        self.notify_trace_mode(error_handler.is_trace_enabled())
        self._cplex_vartype_map = {pyo_ln.VarType.Binary: 'B',
                                   pyo_ln.VarType.Integer: 'I',
                                   pyo_ln.VarType.Continuous: 'C'}
        self._cplex_cttype_map = {pyo_ln.LinearConstraintType.EQ: 'E',
                                   pyo_ln.LinearConstraintType.GE: 'G',
                                   pyo_ln.LinearConstraintType.LE: 'L'}

        self._cplex_objsense_map = {pyo_ln.ObjectiveSense.Minimize: self.__cplex.objective.sense.minimize,
                                    pyo_ln.ObjectiveSense.Maximize: self.__cplex.objective.sense.maximize}


    def _notify_linear_constraint_index(self, linct_idx):
        if linct_idx >= 0:
            self._last_used_ct_index = max(self._last_used_ct_index, linct_idx)

    def _notify_linear_constraint_deleted(self, linct_idx):
        if linct_idx >= 0:
            self._last_used_ct_index -= 1

    def _get_guessed_var_index(self):
        return self._last_used_var_index + 1

    def _notify_var_index(self, idx):
        if idx >= 0:
            self._last_used_var_index = max(self._last_used_var_index, idx)

    def _get_guessed_ct_index(self):
        """
        Makes a guess on the next value of constraint indices,
        when a block with size constraints ha sbeen allocated.

        Note: this comprises both linear constraints and range c onstraints
        :return:
        """
        return self._last_used_ct_index + 1

    def _guessed_ct_index_range(self, block_size):
        first_index = self._last_used_ct_index
        return fast_range(first_index, first_index + block_size)

    @property
    def guess_indices(self):
        """
        Returns true if the engine uses a guessing method for indices.
        If false, always queries indices from CPLEX, but this has an expensive cost,
        amd will not scale well.
        :return:
        """
        return self._index_mode == CplexIndexMode.Guess

    def _set_trace_output(self, ofs):
        cpx = self.__cplex
        cpx.set_log_stream(ofs)
        cpx.set_results_stream(ofs)
        cpx.set_error_stream(ofs)
        cpx.set_warning_stream(ofs)

    def notify_trace_mode(self, is_trace_on):
        assert self.error_handler
        was_printing_cplex_msgs = self.__print_cplex_msgs
        new_printing_cplex_msgs = is_trace_on
        if was_printing_cplex_msgs != new_printing_cplex_msgs:
            if new_printing_cplex_msgs:
               self._set_trace_output(sys.stdout)
            else:
                self._set_trace_output(None)
            self.__print_cplex_msgs = is_trace_on

    def get_var_index(self, dvar):
        dvar_name = dvar.name
        if not dvar_name:
            self.error_handler.fatal("cannot query index for anonymous object: {0!s}", dvar)
        else:
            return self.__cplex.variables.get_indices(dvar_name)

    def get_ct_index(self, ct):
        ctname = ct.name
        if not ctname:
            self.error_handler.fatal("cannot query index for anonymous constraint: {0!s}", ct)
        self_cplex = self.__cplex
        if isinstance(ct, AbstractLinearConstraint):
            return self_cplex.linear_constraints.get_indices(ctname)
        elif isinstance(ct, IndicatorConstraint):
            return self_cplex.indicator_constraints.get_indices(ctname)
        else:
            self.error_handler.fatal("unrecognized constraint to query index: {0!s}", ct)

    @property
    def error_handler(self):
        return self.__errorHandler

    def get_cplex(self):
        """
        Returns the underlying CPLEX object
        :return:
        """
        return self.__cplex

    def getInfinity(self):
        return cplex.infinity

    def _convert_model_value_to_cplex_value(self, type_descr, model_to_cplex_map, model_value):
        """
        :param type_descr: a string e.g. "variable type" describinf the type being mapped
        :param model_to_cplex_map: a dictionary from model values to CPLEX values
        :param model_value: a value from modeling layer, to be translated
        :return:
        """
        res = model_to_cplex_map.get(model_value)
        if res is None:
            self.error_handler.fatal("Unexpected {0}: {1!s}", (type_descr, model_value))
        else:
            return res

    def vartype2cplextype(self, vartype):
        return self._convert_model_value_to_cplex_value("variable type", self._cplex_vartype_map, vartype)

    def _createCplexVarTypeSequence(self, vartype, size=1):
        """ FIXME: Mega Hack here: setting an explicit continuous type
            will lead CPLEX to interpret the problem as a MIP
            got a 1017 error on the production sample for this...
        """
        if vartype is VarType.Continuous:
            return ''
        else:
            return [self.vartype2cplextype(vartype)] * size

    def _model_type_to_cplex_type(self, mtype):
        return 'C' if mtype == pyo_ln.VarType.Continuous else self.vartype2cplextype(mtype)

    def cttype2cplextype(self, cttype):
        return self._convert_model_value_to_cplex_value("constraint type", self._cplex_cttype_map, cttype)

    def objsense2cplexobjsense(self, sense):
        return self._convert_model_value_to_cplex_value("objective sense", self._cplex_objsense_map, sense)

    def _guess_index_range(self, names):
        # take the "real" last index from CPLEX, says 99
        actual_last_index = self.__cplex.variables.get_indices(names[-1])
        # if nb vars = 100, first index is 99-100 + 1= 0 cqfd
        guessed_first_index = actual_last_index - len(names) + 1
        index_range = range(guessed_first_index, actual_last_index + 1)
        return index_range

    # @profile
    def _internal_create_variables(self, allnames, alltypes, alllbs, allubs):
        # INTERNAL
        ret_add = self.__cplex.variables.add(names=allnames, types=alltypes, lb=alllbs, ub=allubs)

        if self._index_mode is CplexIndexMode.UseReturn:
            all_indices = ret_add
        elif self._index_mode is CplexIndexMode.Guess:
            # guess mode.
            all_indices = self._guess_index_range(allnames)
        else:
            # QueryByName mode, safe but verrry slow, no name hashing
            all_indices = self.__cplex.variables.get_indices(allnames)
        self._notify_var_index(all_indices[-1])
        return all_indices

    def createOneVariable(self, vartype, lb, ub, name):
        alltypes = self._createCplexVarTypeSequence(vartype, 1)
        allnames = [name] if name is not None else []
        alllbs = [lb] if lb is not None else []
        allubs = [ub] if ub is not None else []
        ret_val = self.__cplex.variables.add(names=allnames, types=alltypes, lb=alllbs, ub=allubs)
        index_mode = self._index_mode
        if index_mode is CplexIndexMode.UseReturn:
            idx = ret_val[0]
        elif index_mode is CplexIndexMode.Guess:
            idx = self._get_guessed_var_index()
        else:
            idx = self.__cplex.variables.get_indices(name)

        self._notify_var_index(idx)
        return idx

    # @profile
    def createVariables(self, keys, vartype, lbs, ubs, namer):
        if not keys:
            return []
        numberOfVars = len(keys)
        alltypes = self._createCplexVarTypeSequence(vartype, numberOfVars)
        allnames = [namer(k) for k in keys]
        all_indices = self._internal_create_variables(allnames, alltypes, lbs, ubs)
        return all_indices

    def _applyVarFn(self, var, args, setter_fn, getter_fn, transformer=None):
        cpxvars = self.__cplex.variables
        is_var_iterable = is_iterable(var)
        is_arg_iterable = is_iterable(args) and not isinstance(args, str)
        if not is_var_iterable and is_arg_iterable:
            self.error_handler.fatal("Single var requires a numeric argument, not iterable")
        if is_var_iterable:
            it_indices = (_v.safe_index for _v in var)
            it_args = iter(args) if is_arg_iterable else generate_constant(args)
            actual_lbs = []
            for idx in it_indices:
                raw_arg = next(it_args)
                if transformer:
                    applied_arg = transformer(raw_arg)
                else:
                    applied_arg = raw_arg
                setter_fn(cpxvars, idx, applied_arg)
                actual_lbs.append(getter_fn(cpxvars, idx))
            return actual_lbs
        else:
            # newLb assumed to be NOT iterable at this point.
            var_idx = var.safe_index
            applied_arg = args if transformer is None else transformer(args)
            setter_fn(cpxvars, var.safe_index, applied_arg)
            actual_new_arg = getter_fn(cpxvars, var_idx)
            self.error_handler.ensure(applied_arg == actual_new_arg,
                                      "Inconsistent bounds after update, expecting {0!s}, got {1!s}",
                                      applied_arg, actual_new_arg)
            return actual_new_arg

    _getset_map = {"lb": (cplex._internal._subinterfaces.VariablesInterface.set_lower_bounds,
                          cplex._internal._subinterfaces.VariablesInterface.get_lower_bounds),
                   "ub": (cplex._internal._subinterfaces.VariablesInterface.set_upper_bounds,
                          cplex._internal._subinterfaces.VariablesInterface.get_upper_bounds),
                   "name": (cplex._internal._subinterfaces.VariablesInterface.set_names,
                            cplex._internal._subinterfaces.VariablesInterface.get_names)}

    def get_var_attribute(self, dvar, attr_name):
        getset_tuple = self._getset_map.get(attr_name)
        if not getset_tuple:
            self.error_handler.warning("unsupported attribute: {0}", attr_name)
        else:
            getter_fn = getset_tuple[1]
            return getter_fn(self.__cplex.variables, dvar.index)

    def set_var_attribute(self, dvar, attr_name, attr_val):
        getset_tuple = self._getset_map.get(attr_name)
        if not getset_tuple:
            self.error_handler.warning("unsupported attribute: {0}", attr_name)
        else:
            (setter, getter) = getset_tuple
            return self._applyVarFn(dvar, attr_val, setter, getter)

    def setVarName(self, var, new_name):
        return self._applyVarFn(var, new_name,
                                cplex._internal._subinterfaces.VariablesInterface.set_names,
                                cplex._internal._subinterfaces.VariablesInterface.get_names)

    def setVarLb(self, var, newLb):
        return self._applyVarFn(var, newLb,
                                cplex._internal._subinterfaces.VariablesInterface.set_lower_bounds,
                                cplex._internal._subinterfaces.VariablesInterface.get_lower_bounds)

    def setVarUb(self, var, newUb):
        return self._applyVarFn(var, newUb,
                                cplex._internal._subinterfaces.VariablesInterface.set_upper_bounds,
                                cplex._internal._subinterfaces.VariablesInterface.get_upper_bounds)

    def setVarType(self, var, new_type):
        _type_transformer = lambda mtype: self._model_type_to_cplex_type(new_type)
        return self._applyVarFn(var, new_type,
                                cplex._internal._subinterfaces.VariablesInterface.set_types,
                                cplex._internal._subinterfaces.VariablesInterface.get_types,
                                transformer=_type_transformer
                                )

    def getNumberOfColumns(self):
        return self.__cplex.variables.get_num()

    def get_solve_attribute(self, attr, index_seq):
        ''' Returns a sequence of attributes from the engine'''
        self._check_is_solved_ok()

        indices = list(index_seq)
        if attr == "slacks":
            all_attributes = self.__cplex.solution.get_linear_slacks(indices)
        elif attr == "duals":
            all_attributes = self.__cplex.solution.get_dual_values(indices)
        elif attr == "reduced_costs":
            all_attributes = self.__cplex.solution.get_reduced_costs(indices)
        else:
            self.error_handler.error('*unexpected attribute name: {0!s}', attr)
            return {}
        assert len(indices) == len(all_attributes)
        filtered_attr_map = {indices[i]: all_attributes[i] for i in range(len(indices)) if all_attributes[i]}
        return filtered_attr_map

    def _linexpr_to_cplex(self, expr):
        linexpr1 = self._linexpr_to_cplex1(expr)
        return [linexpr1] if linexpr1 else []

    def _linexpr_to_cplex1(self, expr):
        ''' convert a linear expression into a list of two lists (of the same length)
           1. a list of indices taken from previous post to CPLEX
           2. a list of coefs
           note the trick to iterate the expression only once:
           build a zipped list of tuples, then unzip it and return thr list of lists.
        '''
        all_indices = [dvar.get_index() for dvar in expr.iter_variables()]
        if all_indices:
            return [all_indices,  expr._get_coefs()]
        else:
            return []

    # @profile
    def _make_cpx_linexpr(self, linear_ct):
        """
        Builds two lists, one for indices, one for coefs, of variables
        in lef_expr - right_expr.
        Keep the ordering: left variables first, then those from right_expr that are not in left_expr

        :param linear_ct: a linear constraint of type: expr1 OP expr2
        :return: either [] or a list of two lists indices and coefs
        """
        left_expr = linear_ct.left_expr
        right_expr = linear_ct.right_expr
        if right_expr.is_constant():
            all_indices = [dv.get_index() for dv in left_expr.iter_variables()]
            all_coefs = left_expr._get_coefs()
        elif left_expr.is_constant():
            all_indices = [dv.get_index() for dv in right_expr.iter_variables()]
            all_coefs = [-x for x in right_expr._get_coefs()]
        else:
            all_indices = [dv.get_index() for dv in linear_ct.iter_variables()]
            all_coefs   = [left_expr.unchecked_get_coef(dv) - right_expr.unchecked_get_coef(dv) for dv in linear_ct.iter_variables()]
        if all_indices:
            return [all_indices, all_coefs]
        else:
            return []

    def __stop_here(self):
        pass

    # @profile
    def _make_cplex_linear_ct(self, cpx_lin_expr, ctype, rhs, name, scope):
        """
        INTERNAL. fundamental way to post a linear ct to CPLEX.
        :param cpx_lin_expr:
        :param ctype:
        :param rhs:
        :param name:
        :param scope: a name scope with counter
        :return:
        """
        cpx_rhs = [rhs]
        cpx_type = self.cttype2cplextype(ctype)
        cpxnames = [name] if name else []

        linearcts = self.__cplex.linear_constraints
        ret_add = linearcts.add(lin_expr=cpx_lin_expr, senses=cpx_type, rhs=cpx_rhs, names=cpxnames)

        self_index_mode = self._index_mode
        if self_index_mode is CplexIndexMode.Guess:
            cpx_ct_index = self._get_guessed_ct_index()
            self._notify_linear_constraint_index(cpx_ct_index)
        elif self_index_mode is CplexIndexMode.UseReturn:
            cpx_ct_index = ret_add[0]
        else:  # self_index_mode is CplexIndexMode.Query:
            if name:
                cpx_ct_index = linearcts.get_indices(name)
                if cpx_ct_index != self._last_used_ct_index + 1:
                    self.__stop_here()
                self._notify_linear_constraint_index(cpx_ct_index)
            else:
                self.error_handler.trace("Cannot get index for anonymous constraint: {0!s}", ())
                # there is nothing to notify
                return -1

        return cpx_ct_index

    # @profile
    def createBinaryLinearConstraint(self, binaryct):
        cpx_linexp1 = self._make_cpx_linexpr(binaryct)
        # wrap one more time
        cpx_linexp = [cpx_linexp1] if cpx_linexp1 else []
        # returns a number
        num_rhs = binaryct.rhs()
        ct_scope = binaryct.model._get_ct_scope(binaryct)
        return self._make_cplex_linear_ct(cpx_lin_expr=cpx_linexp,
                                          ctype=binaryct.type,
                                          rhs=num_rhs, name=binaryct.name,
                                          scope=ct_scope)

    def create_block_linear_constraints(self, linct_seq):
        block_size = len(linct_seq)
        cpx_rhss   = [ct.rhs() for ct in linct_seq]
        cpx_senses = [self.cttype2cplextype(ct.type) for ct in linct_seq]
        cpx_names  = [ct.name for ct in linct_seq]
        cpx_linexprs = [self._make_cpx_linexpr(ct) for ct in linct_seq]

        cpx_linear = self.__cplex.linear_constraints
        ret_add = cpx_linear.add(lin_expr=cpx_linexprs, senses=cpx_senses, rhs=cpx_rhss, names=cpx_names)

        # -- diffent ways to query the range of new indices ---
        self_index_mode = self._index_mode
        if self_index_mode is CplexIndexMode.Guess:
            cpx_ct_indices = self._guessed_ct_index_range(block_size)
            self._notify_linear_constraint_index(cpx_ct_indices)
        elif self_index_mode is CplexIndexMode.UseReturn:
            cpx_ct_indices = ret_add  # returns a range
        else:  # self_index_mode is CplexIndexMode.Query:
            if cpx_names:
                cpx_ct_indices = cpx_linear.get_indices(cpx_names)
                self._notify_linear_constraint_index(cpx_ct_indices[-1]) # max is last?
            else:
                self.error_handler.trace("Cannot get index for anonymous constraint: {0!s}", ())
                # there is nothing to notify
                return -1

        return cpx_ct_indices

    def createRangeConstraint(self, rngConstraint):
        """
        Post a range constraint to CPLEX
        :param rngConstraint:
        :return:
        """
        linearcts = self.__cplex.linear_constraints
        expr = rngConstraint.expr
        offset = expr.constant
        lhs = rngConstraint.lb
        rhs = rngConstraint.ub
        cpx_lin_expr = self._linexpr_to_cplex(expr)

        cpx_rhs = [rhs - offset]
        cpx_range_values = [lhs - rhs]  # should be negative ???
        cpx_type = 'R'  # TODO: store this
        ctname = rngConstraint.name
        cpxnames = [ctname] if ctname else []
        ret_add = linearcts.add(lin_expr=cpx_lin_expr,
                                senses=cpx_type, rhs=cpx_rhs,
                                range_values=cpx_range_values,
                                names=cpxnames)
        self_index_mode = self._index_mode
        if self_index_mode is CplexIndexMode.Guess:
            cpx_ct_index = self._get_guessed_ct_index()
            self._notify_linear_constraint_index(cpx_ct_index)
        elif self_index_mode is CplexIndexMode.Query:
            if ctname:
                cpx_ct_index = linearcts.get_indices(ctname)
                if cpx_ct_index != self._get_guessed_ct_index():
                    pass
                self._notify_linear_constraint_index(cpx_ct_index)
            else:
                cpx_ct_index = -1
        elif self_index_mode is CplexIndexMode.UseReturn:
            # first element of range
            cpx_ct_index = ret_add[0]
        else:
            cpx_ct_index = -1
            self.error_handler.fatal("Unexpected index mode: {0!s}", self_index_mode)
        return cpx_ct_index

    def createIndicatorConstraint(self, indicator_ct):
        """
        Post an indicator ct to CPLEX
        :param indicator_ct:
        :return:
        """
        self.error_handler.ensure(isinstance(indicator_ct, pyo_ln.IndicatorConstraint), '*Not an indicator constraint',
                                  indicator_ct)
        linear_ct = indicator_ct.linear_constraint
        ct_name = indicator_ct.name
        active_value = 1 - indicator_ct.active_value
        binary_var = indicator_ct.indicator_var
        binary_index = binary_var.safe_index

        # the linear ct is not posted to CPLEX,
        # but we need to convert it to linexpr
        cpx_linexpr = self._make_cpx_linexpr(linear_ct)
        rhs = linear_ct.rhs()
        cpx_name = ct_name or ''
        cpx_sense = self.cttype2cplextype(linear_ct.type)

        cpx_indicators = self.__cplex.indicator_constraints
        cpx_complemented = active_value
        ret_add = cpx_indicators.add(cpx_linexpr, cpx_sense, rhs, binary_index, cpx_complemented, cpx_name)
        if self._index_mode is CplexIndexMode.UseReturn:
            cpx_indicator_index = ret_add  # for indicators, CPLEX returns the index, not a range...
        else:
            cpx_indicator_index = cpx_indicators.get_indices(ct_name) if ct_name else -1
        return cpx_indicator_index

    def remove_constraint(self, ct):
        ct_index = ct.safe_index
        # we have a safe index
        self.__cplex.linear_constraints.delete(ct_index)
        self._notify_linear_constraint_deleted(ct_index)

    def set_objective(self, sense, expr):
        cpx_objective = self.__cplex.objective
        cpx_obj_sense = self.objsense2cplexobjsense(sense)
        # convert expr to a list of pairs (index, coef)
        index_coef_seq = [(dv.get_index(), expr.unchecked_get_coef(dv)) for dv in expr.iter_variables()]
        if index_coef_seq:
            cpx_objective.set_linear(index_coef_seq)
            cpx_objective.set_sense(cpx_obj_sense)

    def clear_objective(self, expr):
        """
        Do not send an empty list otherwise a crash occurs.
        :param expr:
        :return:
        """
        if not expr.is_constant():
            var_zero_seq = [(var.index, 0) for var in expr.iter_variables()]

            self.__cplex.objective.set_linear(var_zero_seq)
            # set_linear() does NOT reset the objective!
            # IndexError: tuple index out of range

    @staticmethod
    def status2string(cpx_status):
        ''' Converts a CPLEX integer status value to a string'''
        return _subinterfaces.SolutionInterface.status.__getitem__(cpx_status)

    def get_status_as_string(self):
        ''' Returns the solve status as a string.'''
        raw_status = self.__cplex.solution.get_status()
        return self.status2string(raw_status)

    @staticmethod
    def isRelaxedStatusOK(status):
        # __CPLEX_RELAX_OK_STATUSES = frozenset([126, 16, 18, 14, 17, 19, 15])
        # list all status values for which there is a relaxed solution.
        # include QUAD values for the future, though for now the modeling doe snot support quads
        __CPLEX_RELAX_OK_STATUSES = {cpx_cst.CPXMIP_OPTIMAL_RELAXED_INF,
                                     cpx_cst.CPXMIP_OPTIMAL_RELAXED_SUM,
                                     cpx_cst.CPXMIP_OPTIMAL_RELAXED_QUAD,
                                     cpx_cst.CPXMIP_FEASIBLE_RELAXED_INF,
                                     cpx_cst.CPXMIP_FEASIBLE_RELAXED_QUAD,
                                     cpx_cst.CPXMIP_FEASIBLE_RELAXED_SUM,
                                     cpx_cst.CPX_STAT_FEASIBLE_RELAXED_SUM,
                                     cpx_cst.CPX_STAT_FEASIBLE_RELAXED_INF,
                                     cpx_cst.CPX_STAT_FEASIBLE_RELAXED_QUAD
                                     }
        return status in __CPLEX_RELAX_OK_STATUSES

    __CPLEX_SOLVE_OK_STATUSES = {1,    # CPX_STAT_OPTIMAL
                                 6,    # CPX_STAT_NUM_BEST: solution exists but numerical issues
                                 101,  # CPXMIP_OPTIMAL
                                 102,  # CPXMIP_OPTIMAL_TOL
                                 104,  # CPXMIP_SOL_LIM
                                 105,  # CPXMPI_NODE_LIM_FEAS
                                 107,  # CPXMIP_TIME_LIM_FEAS
                                 109,  # CPXMIP_FAIL_FEAS : what is this ??
                                 111,  # CPXMIP_MEM_LIM_FEAS
                                 113,  # CPXMIP_ABORT_FEAS
                                 116,  # CPXMIP_FAIL_FEAS_NO_TREE : integer sol exists (????)
                                 129,  # CPXMIP_OPTIMAL_POPULATED
                                 130   # CPXMIP_OPTIMAL_POPULATED_TOL
                                 }

    @staticmethod
    def is_solve_status_ok(status):
        # Converts a raw CPLEX status to a boolean
        return status in CplexEngine.__CPLEX_SOLVE_OK_STATUSES

    def can_solve(self):
        return True

    def name(self):
        return 'cplex'

    def solve(self, mdl):
        cpx = self.__cplex
        pb_stats = cpx.get_stats()
        self.__solveCount += 1
        # print("--> starting CPLEX solve #", self.__solveCount)
        cpx_status_string = None
        try:
            mip_start = mdl.mip_start
            if mip_start and mip_start.check_as_mip_start():
                # convert explicit values as tuples of (index, value)
                effort_level = cpx.MIP_starts.effort_level.repair
                all_indices = [dv.get_index() for dv, _ in mip_start.iter_var_values()]
                all_values = [val for _, val in mip_start.iter_var_values()]
                cpx.MIP_starts.add([all_indices, all_values], effort_level)
                pass
            cpx.solve()  # returns nothing in Python
            cpx_status = cpx.solution.get_status()
            cpx_status_string = self.__cplex.solution.get_status_string(cpx_status)
            solve_ok = self.is_solve_status_ok(cpx_status)

        except cplex.exceptions.CplexSolverError as cpx_s:
            self.error_handler.error("CPLEX Error: {0!s}, code={1}",
                                     (cpx_s.args[0], cpx_s.args[2])) # tuples required here...
            solve_ok = False

        except cplex.exceptions.CplexError as cpx_e:
            self.error_handler.error("CPLEX error: {0}", cpx_e.message)
            solve_ok = False

        self.__lastSolveStatus = solve_ok
        new_solution = None
        if solve_ok:
            # cpompute correct objective including constant term
            full_obj = cpx.solution.get_objective_value() + mdl.objective_expr.constant
            self.__lastObjectiveValue = full_obj
            # we need to build this list (maybe cache it?)
            all_var_indices = [dvar.index for dvar in mdl.iter_variables()]
            if all_var_indices:
                # do not query values on an empty model...
                all_var_values = self.__cplex.solution.get_values(all_var_indices)
                var_value_map = dict(zip(mdl.iter_variables(), all_var_values))
            else:
                var_value_map = {}

            new_solution = SolveSolution(mdl, obj=full_obj,
                                         var_value_map=var_value_map,
                                         engine_name=self.name(),
                                         keep_zeros=False,
                                         rounding=True)
            new_solution._set_solve_status(self.get_solve_status())
            # cache attributes?
        else:
            mdl.notify_solve_failed()
        if cpx_status_string:
            mdl.error_handler.trace("CPLEX solve returns with status: {0}", (cpx_status_string,))
        return new_solution

    def _check_is_solved_ok(self):
        """
        INTERNAL: checks the engine has recently been solved ok.
        Either raise an exception or returns None.
        :return:
        """
        self.error_handler.ensure(self.__solveCount > 0, "Model is not solved yet")
        self.error_handler.ensure(self.__lastSolveStatus, "Last solve failed")

    def get_solutions(self, dvars):
        # self._check_is_solved_ok()
        if not dvars:
            return {}

        indices = []
        for w in dvars:
            if is_number(w):
                indices.append(w)
            elif isinstance(w, pyo_ln.Var):
                indices.append(w.index)
            else:
                self.error_handler.fatal("expecting index or variable, got: {0!s}", w)

        all_values = self.__cplex.solution.get_values(indices)
        return dict(zip(indices, all_values))

    def solve_relaxed(self, relaxable_groups, optimize, limits):
        """ Runs feasopt with a set of relaxable cts and numerical  preferences.

        Args:
            relaxable_groups:
            optimize: True if the model has a non-numeric objective
            limits: a tuple with limits on the execution of each pass.

        Returns:
            a solution object, or None.
        """
        relax_gap, relax_max_nb_sol, relax_pass_time_limit = limits
        cplex_params = self.__cplex.parameters
        cplex_feasopt_mode_param = cplex_params.feasopt.mode
        # which forced mode for feasopt? switch this flag for testing
        # with 12.67.2 and nurse, INF is very slow...
        use_sum_or_nb = True
        if use_sum_or_nb:
            if optimize:
                new_mode = cplex_params.feasopt.mode.values.opt_sum
            else:
                new_mode = cplex_params.feasopt.mode.values.min_sum
        else:
            if optimize:
                new_mode = cplex_params.feasopt.mode.values.opt_inf
            else:
                new_mode = cplex_params.feasopt.mode.values.min_inf

        overwritten_params = {cplex_feasopt_mode_param: new_mode}
        if relax_gap > 0:
            overwritten_params[cplex_params.mip.tolerances.mipgap] = relax_gap
        if relax_max_nb_sol >= 1:
            overwritten_params[cplex_params.mip.limits.solutions] = relax_max_nb_sol
        if relax_pass_time_limit >= 1:  # no less than 1s.
            overwritten_params[cplex_params.timelimit] = relax_pass_time_limit

        class cpx_ovwerwrite_params_execution_context(object):
            # internal context manager to handle forcing parameters during relaxation.

            def __init__(self, cplex_to_overwrite, overwrite_param_dict):
                assert isinstance(overwrite_param_dict, dict)
                self._cplex = cplex_to_overwrite
                self._overwrite_param_dict = overwrite_param_dict
                # store current values
                self._saved_param_values = {p: p.get() for p in overwrite_param_dict}
                # for p, v in iteritems(overwrite_param_dict):
                # print("-- overwriting parameter {0} with value: {1!s}".format(p._name, v))

            def __enter__(self):
                # force overwrite values.
                for p, v in iteritems(self._overwrite_param_dict):
                    p.set(v)
                # return the Cplex instance with the overwritten parameters.
                return self._cplex

            def __exit__(self, exc_type, exc_val, exc_tb):
                # whatever happened, restore saved parameter values.
                for param, old_value in iteritems(self._saved_param_values):
                    param.set(old_value)

        cpx_relax_groups = []
        cpx = self.__cplex
        with cpx_ovwerwrite_params_execution_context(cpx, overwritten_params) as cpx2:
            cpx_feasopt = cpx2.feasopt
            all_indices = []
            all_cts = []
            for (pref, group_cts) in relaxable_groups:
                self.error_handler.ensure(is_number(pref) and pref > 0,
                                          "Relaxation preference must be strict positive number", pref)
                self.error_handler.ensure(is_iterable(group_cts), "Non-iterable relaxable cts", group_cts)
                if not group_cts:
                    continue

                group_indices = [ct.safe_index for ct in group_cts]
                cpx_relax_groups.append(cpx_feasopt.linear_constraints(pref, group_indices))
                all_indices.append(group_indices)
                all_cts.append(group_cts)
                # at this stage, we have a list of groups
                # each group is itself a list
                # the first item is a number, th epreference
                # the second item is a list of indices.
                cpx_feasopt(*cpx_relax_groups)

        # feasopt state is restored by now
        cpx_solution = cpx.solution
        feas_status = cpx_solution.get_status()
        feas_ok = self.isRelaxedStatusOK(feas_status)
        feas_obj = cpx_solution.get_objective_value() if feas_ok else 0
        # feas_status_string = CplexEngine.status2string(feas_status)
        # print("* feasopt returns with status: %s, obj=%g"%(feas_status_string, feas_obj))
        return feas_ok, feas_obj

    def getInfeasibilities(self, cts):
        indices = [ct.index for ct in cts]
        # PCO: Daniel Junglas confirms using [] uses the last solution vector
        return self.__cplex.solution.infeasibility.linear_constraints([], indices)

    def dump(self, path):
        if path.find('.') > 0:
            self.__cplex.write(path)
        else:
            self.__cplex.write(path, filetype="lp")

    def getSolutionValues(self, varSequence):
        return self.__cplex.solution.get_values(varSequence)

    def get_problem_type(self):
        """ CPLEX wrapper returns an integer."""
        return self.__cplex.get_problem_type()

    def end(self):
        """ terminate the engine, cannot find this in the doc.
        """
        del self.__cplex
        self.__cplex = None

    # noinspection PyProtectedMember
    def isMIP(self):
        cpx = self.__cplex
        cpx._is_MIP()
        _all_mip_problems = frozenset({'MIP', 'MILP', 'fixedMILP', 'MIQP', 'fixedMIQP'})
        cpx_problem_type = cpx.problem_type[cpx.get_problem_type()]
        return cpx_problem_type in _all_mip_problems

    def connect_progress_listeners(self, progress_listener_list):
        if not progress_listener_list:
            self.error_handler.info("Null progress listeners", None)
        elif not self.__cplex._is_MIP():
            self.error_handler.warning("Not a MIP problem, cannot use progress listeners")
        else:
            ccb = self.__cplex.register_callback(ConnectListenersCallback)
            ccb.initialize(progress_listener_list)

    def set_parameter(self, parameter, value):
        # value check is up to the caller.
        self.__cplex._env.parameters._set(parameter.cpx_id, value)

    def get_parameter(self, parameter):
        try:
            return self.__cplex._env.parameters._get(parameter.cpx_id)
        except CplexError:
            return parameter.default_value

    def get_solve_status(self):
        from docloud.status import JobSolveStatus
        # In this function we try to do the exact same mappings as the IloCplex C++ and Java classes.
        # However, this is not always possible since the C++ and Java implementations are not consistent
        # and sometimes they are even in error (see RTC-21923).
        cpx_status = self.__cplex.solution.get_status()
        if cpx_status in {cpx_cst.CPXMIP_ABORT_FEAS,
                          cpx_cst.CPXMIP_DETTIME_LIM_FEAS,
                          cpx_cst.CPXMIP_FAIL_FEAS,
                          cpx_cst.CPXMIP_FAIL_FEAS_NO_TREE,
                          cpx_cst.CPXMIP_MEM_LIM_FEAS,
                          cpx_cst.CPXMIP_NODE_LIM_FEAS,
                          cpx_cst.CPXMIP_TIME_LIM_FEAS
                          }:
            return JobSolveStatus.FEASIBLE_SOLUTION

        elif cpx_status in {cpx_cst.CPXMIP_ABORT_INFEAS,
                            cpx_cst.CPXMIP_DETTIME_LIM_INFEAS,
                            cpx_cst.CPXMIP_FAIL_INFEAS,
                            cpx_cst.CPXMIP_FAIL_INFEAS_NO_TREE,
                            cpx_cst.CPXMIP_MEM_LIM_INFEAS,
                            cpx_cst.CPXMIP_NODE_LIM_INFEAS,
                            cpx_cst.CPXMIP_TIME_LIM_INFEAS
                            }:
            # Hit a limit without a feasible solution: We don't know anything about the solution.
            return JobSolveStatus.UNKNOWN
        elif cpx_status in {cpx_cst.CPXMIP_OPTIMAL,
                            cpx_cst.CPXMIP_OPTIMAL_TOL}:
            return JobSolveStatus.OPTIMAL_SOLUTION
        elif cpx_status is cpx_cst.CPXMIP_SOL_LIM:
            #  return hasSolution(env, lp) ? JobSolveStatus.FEASIBLE_SOLUTION : JobSolveStatus.UNKNOWN;
            return JobSolveStatus.FEASIBLE_SOLUTION
        elif cpx_status is cpx_cst.CPXMIP_INForUNBD:
            return JobSolveStatus.INFEASIBLE_OR_UNBOUNDED_SOLUTION
        elif cpx_status in {cpx_cst.CPXMIP_UNBOUNDED,
                            cpx_cst.CPXMIP_ABORT_RELAXATION_UNBOUNDED}:
            return JobSolveStatus.UNBOUNDED_SOLUTION
        elif cpx_status is cpx_cst.CPXMIP_INFEASIBLE:  # proven infeasible
            return JobSolveStatus.INFEASIBLE_SOLUTION
        elif cpx_status is cpx_cst.CPXMIP_OPTIMAL_INFEAS:  # optimal with unscaled infeasibilities
            # TODO: What exactly do we return here? There is an optimal solution but that solution is
            # infeasible after unscaling.
            return JobSolveStatus.OPTIMAL_SOLUTION

        #  feasopt status values
        elif cpx_status in frozenset({
            cpx_cst.CPXMIP_ABORT_RELAXED,  # relaxed solution is available and can be queried
            cpx_cst.CPXMIP_FEASIBLE  # problem feasible after phase I and solution available
        }):
            return JobSolveStatus.FEASIBLE_SOLUTION
        elif cpx_status in {cpx_cst.CPXMIP_FEASIBLE_RELAXED_INF,
                            cpx_cst.CPXMIP_FEASIBLE_RELAXED_QUAD,
                            cpx_cst.CPXMIP_FEASIBLE_RELAXED_SUM
                            }:
            return JobSolveStatus.UNKNOWN
        elif cpx_status in {cpx_cst.CPXMIP_OPTIMAL_RELAXED_INF,
                            cpx_cst.CPXMIP_OPTIMAL_RELAXED_QUAD,
                            cpx_cst.CPXMIP_OPTIMAL_RELAXED_SUM
                            }:
            return JobSolveStatus.INFEASIBLE_SOLUTION

        # populate status values
        elif cpx_status in {cpx_cst.CPXMIP_OPTIMAL_POPULATED
                            # ,cpx_cst.CPXMIP_OPTIMAL_POPULATED_TO
                            }:
            return JobSolveStatus.OPTIMAL_SOLUTION
        elif cpx_status is cpx_cst.CPXMIP_POPULATESOL_LIM:
            # minimal value for CPX_PARAM_POPULATE_LIM is 1! So there must be a solution
            return JobSolveStatus.FEASIBLE_SOLUTION

        elif cpx_status is cpx_cst.CPX_STAT_OPTIMAL:
            return JobSolveStatus.OPTIMAL_SOLUTION

        elif cpx_status is cpx_cst.CPX_STAT_INFEASIBLE:
            return JobSolveStatus.INFEASIBLE_SOLUTION

        # cpx_cst.CPX_STAT_ABORT_USER:
        # cpx_cst.CPX_STAT_ABORT_DETTIME_LIM:
        # cpx_cst.CPX_STAT_ABORT_DUAL_OBJ_LIM:
        # cpx_cst.CPX_STAT_ABORT_IT_LIM:
        # cpx_cst.CPX_STAT_ABORT_PRIM_OBJ_LIM:
        # cpx_cst.CPX_STAT_ABORT_TIME_LIM:
        #   switch (primalDualFeasible(env, lp)) {
        #   case PRIMAL_FEASIBLE: return JobSolveStatus.FEASIBLE_SOLUTION
        #   case PRIMAL_DUAL_FEASIBLE: return JobSolveStatus.OPTIMAL_SOLUTION
        #   case DUAL_FEASIBLE: return JobSolveStatus.UNKNOWN;
        #   default: return JobSolveStatus.UNKNOWN;
        #   }
        #
        # cpx_cst.CPX_STAT_ABORT_OBJ_LIM:
        #   /** FIXME: Our Java API returns ERROR here while the C++ API returns Feasible if primal feasible
        #    *         and Unknown otherwise. Since we don't have ERROR in IloSolveStatus we emulate the
        #    *         C++ behavior (this is more meaningful anyway). In the long run we should make sure
        #    *         all the APIs behave in the same way.
        #    */
        #   switch (primalDualFeasible(env, lp)) {
        #   case PRIMAL_FEASIBLE: return JobSolveStatus.FEASIBLE_SOLUTION
        #   case PRIMAL_DUAL_FEASIBLE: return JobSolveStatus.FEASIBLE_SOLUTION
        #   default: return JobSolveStatus.UNKNOWN;
        #   }
        #
        # cpx_cst.CPX_STAT_FIRSTORDER:
        #   // See IloCplexI::CplexToAlgorithmStatus()
        #   return primalFeasible(env, lp) ? JobSolveStatus.FEASIBLE_SOLUTION : JobSolveStatus.UNKNOWN;

        elif cpx_status is cpx_cst.CPX_STAT_CONFLICT_ABORT_CONTRADICTION:
            # Numerical trouble in conflict refiner.
            #  FIXME: C++ and Java both return Error here although a conflict is
            #          available (but nor proven to be minimal). This looks like a bug
            #          since no exception is thrown there. In IloSolveStatus we don't
            #          have ERROR, so we return UNKNOWN instead. This is fine for now
            #          since we do not support the conflict refiner anyway.
            #
            return JobSolveStatus.UNKNOWN

        elif cpx_status in {

            cpx_cst.CPX_STAT_CONFLICT_ABORT_DETTIME_LIM,
            cpx_cst.CPX_STAT_CONFLICT_ABORT_IT_LIM,
            cpx_cst.CPX_STAT_CONFLICT_ABORT_MEM_LIM,
            cpx_cst.CPX_STAT_CONFLICT_ABORT_NODE_LIM,
            cpx_cst.CPX_STAT_CONFLICT_ABORT_OBJ_LIM,
            cpx_cst.CPX_STAT_CONFLICT_ABORT_TIME_LIM,
            cpx_cst.CPX_STAT_CONFLICT_ABORT_USER
        }:
            # /** FIXME: C++ and Java return Error here. This is almost certainly wrong.
            # *         Docs say "a conflict is available but not minimal".
            #  *         This is particularly erroneous if no exception gets thrown.
            #  *         See RTC-21923.
            #  *         In IloSolveStatus we don't have ERROR, so we return UNKNOWN instead.
            #  *         This should not be a problem since right now we don't support the
            #  *         conflict refiner anyway.
            #  */
            return JobSolveStatus.UNKNOWN
        elif cpx_status is cpx_cst.CPX_STAT_CONFLICT_FEASIBLE:
            return JobSolveStatus.FEASIBLE_SOLUTION
        elif cpx_status is cpx_cst.CPX_STAT_CONFLICT_MINIMAL:
            return JobSolveStatus.INFEASIBLE_SOLUTION
        elif cpx_status in {cpx_cst.CPX_STAT_FEASIBLE_RELAXED_INF,
                            cpx_cst.CPX_STAT_FEASIBLE_RELAXED_QUAD,
                            cpx_cst.CPX_STAT_FEASIBLE_RELAXED_SUM,
                            }:
            return JobSolveStatus.UNKNOWN

        elif cpx_status is cpx_cst.CPX_STAT_FEASIBLE:
            return JobSolveStatus.FEASIBLE_SOLUTION
        elif cpx_status in {cpx_cst.CPX_STAT_OPTIMAL_RELAXED_INF,
                            cpx_cst.CPX_STAT_OPTIMAL_RELAXED_QUAD,
                            cpx_cst.CPX_STAT_OPTIMAL_RELAXED_SUM}:
            return JobSolveStatus.INFEASIBLE_SOLUTION

        elif cpx_status is cpx_cst.CPX_STAT_NUM_BEST:
            #  Solution available but not proved optimal (due to numeric difficulties)
            # assert(hasSolution(env, lp));
            return JobSolveStatus.UNKNOWN

        elif cpx_status is cpx_cst.CPX_STAT_OPTIMAL_INFEAS:  # infeasibilities after unscaling
            # assert(hasSolution(env, lp));
            return JobSolveStatus.OPTIMAL_SOLUTION

        elif cpx_status is cpx_cst.CPX_STAT_INForUNBD:  # Infeasible or unbounded in presolve.
            return JobSolveStatus.INFEASIBLE_OR_UNBOUNDED_SOLUTION
        elif cpx_status is cpx_cst.CPX_STAT_OPTIMAL_FACE_UNBOUNDED:
            #    unbounded optimal face (barrier only)
            # // CPX_STAT_OPTIMAL_FACE_UNBOUNDED is explicitly an error in Java and implicitly (fallthrough)
            # // an error in C++. So it should be fine to produce an error here as well.
            # // In IloSolveStatus we don't have ERROR, so we return UNKNOWN instead.
            # // In case of ERROR we should have seen a non-zero status anyway and the
            # // user should not care too much about the returned status.
            return JobSolveStatus.UNKNOWN
        elif cpx_status is cpx_cst.CPX_STAT_UNBOUNDED:
            # definitely unbounded
            return JobSolveStatus.UNBOUNDED_SOLUTION
        else:
            return JobSolveStatus.UNBOUNDED_SOLUTION


@contextmanager
def overload_cplex_parameter_values(cpx_engine, overload_dict):
    old_values = {p: p.get() for p in overload_dict}
    try:
        yield cpx_engine
    finally:
        # restore params
        for p, saved_value in iteritems(old_values):
            p.set(saved_value)
