# pylint: disable=too-many-lines
from __future__ import print_function

from itertools import product

from docplex.mp.lp_printer import LPModelPrinter
from docplex.mp.mprinter import ModelPrettyPrinter
from docplex.mp.environment import get_configured_env
from docplex.mp.error_handler import DefaultErrorHandler, InfoLevel
from docplex.mp.progress import ProgressListener
from docplex.mp.params.parameter_hierarchy import root_params_make
from docloud.status import JobSolveStatus
from docplex.mp.docloud_engine import DOcloudEngine

# from docplex.Utils import MyTimer


from docplex.mp.linear import *
from docplex.mp.utils import fast_range

from docplex.mp.engine_factory import EngineFactory
from docplex.mp.format import LP_format

from docplex.mp.kpi import DecisionKPI

from docplex.mp.params.parameters import Parameter

import docplex.mp.worker.worker_env as wk_env


class _SolveHook(object):
    # INTERNAL
    def __init__(self):
        pass  # pragma : no cover

    def notify_start_solve(self, attributes):
        """ Notifies the start of a solve.

        Args:
            attributes: A dictionary of string->values with various data attributes of the model.

        """
        pass  # pragma : no cover

    def notify_end_solve(self, has_solution, status, obj, var_value_dict):
        """ Notifies the end of a solve.

        Args:
            has_solution: Boolean, True if solve returned a solution.
            status: An enumerated value of type JobSolveStatus.
            obj: The objective value if solved ok, else irrelevant.
            attributes: A dictionary of variable names to values in the solution.
        """
        pass  # pragma : no cover


class TraceSolveHook(_SolveHook):
    def notify_start_solve(self, attributes):
        print("-> start solve")
        for k, v in iteritems(attributes):
            print("   {0:s} = {1!s}".format(k, v))

    def notify_end_solve(self, has_solution, status, obj, var_value_dict):
        if has_solution:
            print("<- solve succeeds, status={0}, obj={1}".format(status, obj))
            for vn, vv in iteritems(var_value_dict):
                print("  - var \"{0:s}\" = {1!s}".format(vn, vv))
        else:
            print("<- solve fails, status={0}".format(status))


class _ToleranceScheme(object):
    """
        INTERNAL
        """

    def __init__(self, absolute=1e-6, relative=1e-4):
        assert absolute >= 0
        assert relative >= 0
        self._absolute_tolerance = absolute
        self._relative_tolerance = relative

    def compute_tolerance(self, obj):
        abs_obj = abs(obj)
        return max(self._absolute_tolerance, self._relative_tolerance * abs_obj)

    def to_string(self):
        return "tolerance(abs=%f,rel=%f".format(self._absolute_tolerance, self._relative_tolerance)

    def __str__(self):
        return self.to_string()  # pragma: no cover


class _SolveAttribute(object):
    # A generic descriptor class for engine solve attributes (e.g. dual values, reduced costs).

    def __init__(self, name, is_for_vars, requires_solve=True):
        self._name = name
        self._is_for_vars = is_for_vars
        self._requires_solved = requires_solve

    @property
    def name(self):
        return self._name

    @property
    def is_for_vars(self):
        return self._is_for_vars

    @property
    def requires_solved(self):
        return self._requires_solved


class _VariableContainer(object):
    def __init__(self, model, vartype, key_seq, lb, ub, name):
        self._model = model
        self._vartype = vartype
        self._keys = key_seq
        self._lb = lb
        self._ub = ub
        self._namer = name

    def copy(self, target_model):
        copied_ctn = _VariableContainer(target_model, self.vartype, self._keys, self.lb, self.ub, self._namer)
        return copied_ctn

    @property
    def vartype(self):
        return self._vartype

    @property
    def nb_dimensions(self):
        return len(self._keys)

    @property
    def namer(self):
        return self._namer

    @property
    def lb(self):
        return self._lb

    @property
    def ub(self):
        return self._ub

    @property
    def name(self):
        """
        Try to extract a name string from a naming function.
        :return: A string or None.
        """
        if isinstance(self._namer, str):
            raw_name = self._namer
            # drop opl-style formats
            raw_name = raw_name.replace("({%s})", "")
            # purge fields
            pos_pct = raw_name.find('%')
            if pos_pct >= 0:
                return raw_name[:pos_pct - 1]
            elif raw_name.find('{') > 0:
                pos = raw_name.find('{')
                return raw_name[:pos - 1]
            else:
                return raw_name

    def size(self, dim_index):
        return len(self._keys[dim_index]) if dim_index < self.nb_dimensions else 0

    @property
    def dimension_string(self):
        dim_sizes = [self.size(d) for d in range(self.nb_dimensions)]
        dim_string = "".join(["[%d]" % sz for sz in dim_sizes])
        return dim_string

    def to_string(self):
        # dvar xxx
        dim_string = self.dimension_string
        ctname = self._namer or 'x'
        return "dvar {0} {1} {2}".format(self.vartype.short_name, ctname, dim_string)

    def __str__(self):
        return self.to_string()


class _SymbolGenerator(object):
    """
    INTERNAL class
    """

    def __init__(self, pattern):
        ''' Initialize the counter and the pattern.
            Fixes the pattern by suffixing '%d' if necessary.
        '''
        self.__pattern = pattern
        # add offset to counter. this is to avoid 0 in names
        self.__offset = 1
        if pattern.endswith('%d'):
            self.__pattern = pattern
        else:
            self.__pattern = pattern + '%d'
        self._last_index = -1

    def reset(self):
        self._last_index = -1

    def last_index(self):
        return self._last_index

    def notify_new_index(self, new_index):
        """
        Notifies that a new index has been allocated.
        :param new_index:
        :return:
        """
        self._last_index = max(self._last_index, new_index)

    def new_symbol(self):
        """
        Generates and returns a new symbol.
        Guess a new (yet) unallocated index, then use the pattern.
        Note that we use the offset of 1 to generate the name so x1 has index 0, x3 has index 2, etc.
        :return: A symbol string, suposedly not yet allocated.
        """
        guessed_index = self.last_index() + 1
        coined_symbol = self.__pattern % (guessed_index + self.__offset)
        self.notify_new_index(guessed_index)
        return coined_symbol


class ModelStatistics(object):
    """ModelStatistics()

    This class gathers statistics from the model.

    Instances of this class are returned by method 'get_statistics()' on a Model.

    The class contains counters on the various types of variables and constraints
    in the model.

    """

    def __init__(self):
        self._number_of_binary_variables = 0
        self._number_of_integer_variables = 0
        self._number_of_continuous_variables = 0
        self._number_of_le_constraints = 0
        self._number_of_ge_constraints = 0
        self._number_of_eq_constraints = 0
        self._number_of_range_constraints = 0
        self._number_of_indicator_constraints = 0

    def equal_stats(self, other):
        if not isinstance(other, ModelStatistics):
            return False

        if self.number_of_binary_variables != other.number_of_binary_variables:
            return False
        if self.number_of_integer_variables != other.number_of_integer_variables:
            return False
        if self.number_of_continuous_variables != other.number_of_continuous_variables:
            return False
        if self.number_of_linear_constraints != other.number_of_linear_constraints:
            return False
        if self.number_of_le_constraints != other.number_of_le_constraints:
            return False
        if self.number_of_eq_constraints != other.number_of_eq_constraints:
            return False
        if self.number_of_ge_constraints != other.number_of_ge_constraints:
            return False
        if self.number_of_range_constraints != other.number_of_range_constraints:
            return False
        if self._number_of_indicator_constraints != other._number_of_indicator_constraints:
            return False
        # ok all counts are ok.
        return True

    def __eq__(self, other):
        return self.equal_stats(other)

    @staticmethod
    def _make_new_stats(mdl):
        # INTERNAL
        stats = ModelStatistics()
        vartype_count = Counter(dv.vartype for dv in mdl.iter_variables())
        stats._number_of_binary_variables = vartype_count[VarType.Binary]
        stats._number_of_integer_variables = vartype_count[VarType.Integer]
        stats._number_of_continuous_variables = vartype_count[VarType.Continuous]

        linct_count = Counter(ct.type for ct in mdl.iter_binary_constraints())
        stats._number_of_le_constraints = linct_count[LinearConstraintType.LE]
        stats._number_of_eq_constraints = linct_count[LinearConstraintType.EQ]
        stats._number_of_ge_constraints = linct_count[LinearConstraintType.GE]
        stats._number_of_range_constraints = mdl.number_of_range_constraints
        stats._number_of_indicator_constraints = mdl.number_of_indicator_constraints
        return stats

    @property
    def number_of_variables(self):
        """ Returns the total number of variables in the model.

        """
        return self._number_of_binary_variables + \
               self._number_of_integer_variables + \
               self._number_of_continuous_variables

    @property
    def number_of_binary_variables(self):
        """ Returns the number of bianry variables in the model.

        """
        return self._number_of_binary_variables

    @property
    def number_of_integer_variables(self):
        """ Returns the number of integer variables in the model.

        """
        return self._number_of_integer_variables

    @property
    def number_of_continuous_variables(self):
        """ Returns the number of continuous variables in the model.

        """
        return self._number_of_continuous_variables

    @property
    def number_of_linear_constraints(self):
        """ Returns the total number of linear constraints in the model.

        This number comprises all relational constraints: <=, == and >=
        and also range constraints.

        """
        return self._number_of_eq_constraints +\
               self._number_of_le_constraints +\
               self._number_of_ge_constraints +\
               self._number_of_range_constraints

    @property
    def number_of_le_constraints(self):
        """ Returns the number of <= constraints

        """
        return self._number_of_le_constraints

    @property
    def number_of_eq_constraints(self):
        """ Returns the number of == constraints

        """
        return self._number_of_eq_constraints

    @property
    def number_of_ge_constraints(self):
        """ Returns the number of >= constraints

        """
        return self._number_of_ge_constraints

    @property
    def number_of_range_constraints(self):
        """ Returns the number of range constraints.

        Range constraints are of the form L <= expression <= U

        See Also:
            :class:`docplex.mp.linear.RangeConstraint`

        """
        return self._number_of_range_constraints

    def number_of_indicator_constraints(self):
        """ Returns the number of indicator constraints.


        See Also:
            :class:`docplex.mp.linear.IndicatorConstraint`

        """
        return self._number_of_indicator_constraints

    def print_information(self):
        """ Prints model statistics in readable format.

        """
        print(" - number of variables: %d" % self.number_of_variables)
        print("   - binary={0}, integer={1}, continuous={2}".
              format(self.number_of_binary_variables,
                     self.number_of_integer_variables,
                     self.number_of_continuous_variables
                     ))

        print(" - number of constraints: %d" % self.number_of_linear_constraints)
        print(" -   LE={0}, EQ={1}, GE={2}, RNG={3}"
              .format(self._number_of_le_constraints,
                      self._number_of_eq_constraints,
                      self._number_of_ge_constraints,
                      self._number_of_range_constraints))
        if self._number_of_indicator_constraints:
            print(" - number of indicator constraints: %d" % self._number_of_indicator_constraints)

    def __repr__(self):
        return "docplex.mp.Model.ModelStatistics()"


# noinspection PyProtectedMember
class Model(object):
    """ This is the main class to embed modeling objects.

    The Model class acts as a factory to create optimization objects,
    variables, and constraints.
    It provides various accessors and iterators to the modeling objects.
    It also manages solving operations and solution management.


    Args:
        name (str): The name of the model.
        output_level: An instance of enum InfoLevel; default is INFO.
        docloud_context: An instance of DOcloudContext to provide credentials for solving on DOcloud.

    Attributes:
        parameters: Contains the list of CPLEX parameters with their current values.

    Note:

    """

    @staticmethod
    def binary_vartype():
        return VarType.Binary

    @staticmethod
    def integer_vartype():
        return VarType.Integer

    @staticmethod
    def continuous_vartype():
        return VarType.Continuous

    @property
    def environment(self):
        # for a closed model with no CPLEX, numpy, etc return ClosedEnvironment
        # return get_no_graphics_env()
        # from docplex.environment import ClosedEnvironment
        # return ClosedEnvironment
        return get_configured_env()

    def _make_key_seq(self, keys):
        # INTERNAL Takes as input a candidate keys input and returns a valid key sequence
        if is_iterable(keys):
            if has_len(keys):
                return keys
            elif is_iterator(keys):
                return list(keys)
            else:
                # TODO: make a test for this case.
                self.fatal("Cannot handle iterable var keys: {0!s} : no len() and not an iterator",
                           keys)  # pragma : no cover

        elif is_int(keys) and keys >= 0:
            return range(0, keys)
        else:
            self.fatal("Unexpected var keys: {0!s}, expecting either iterable or integer", keys)

    # ---- type checking

    def typecheck_iterable(self, arg):
        # INTERNAL: checks for an iterable
        self.error_handler.ensure(is_iterable(arg), "Expecting iterable, got: {0!s}", arg)

    # safe checks.
    def typecheck_valid_index(self, arg):
        self.__error_handler.ensure(ModelingObject.is_valid_index(arg), "Not an index: {0!s}", arg)

    def typecheck_vartype(self, arg):
        # INTERNAL: check for a valid vartype
        if arg not in {VarType.Binary, VarType.Integer, VarType.Continuous}:
            self.error_handler.fatal("Not a variable type: {0!s}, expecting VarType.Binary,Integer,Continuous", arg)
        return True

    def typecheck_var(self, obj):
        # INTERNAL: check for Var instance
        checked_is_var = isinstance(obj, Var)
        if not checked_is_var:
            self.fatal("Expecting decision variable, got: {0!s} type: {1!s}", obj, type(obj))

    def typecheck_expr(self, arg):
        self.error_handler.ensure(isinstance(arg, Expr), "Expected expression, got: {0!s}", arg)

    def typecheck_constraint(self, obj, ct_subtype=None):
        checked_type = ct_subtype or AbstractConstraint
        checked = isinstance(obj, checked_type)
        self.__error_handler.ensure(checked, "Expecting constraint of type: {0!s}, got: {1!s} with type: {2!s}",
                                    checked_type, obj, type(obj))

    def typecheck_zero_or_one(self, arg):
        if arg != 0 and arg != 1:
            self.fatal("expecting 0 or 1, got: {0!s}", arg)

    def typecheck_int(self, arg):
        self.error_handler.ensure(is_int(arg), "Expecting integer, got: {0!s}", arg)

    def typecheck_num(self, arg, caller=None):
        if caller:
            self.error_handler.ensure(is_number(arg), "{0}: Expecting number, got: {1!s}", caller, arg)
        else:
            self.error_handler.ensure(is_number(arg), "Expecting number, got: {0!s}", arg)

    # --- other checks
    def _check_in_model(self, mobj, typename):
        # produces message of the type: "constraint ... does not belong to model
        if mobj.model != self:
            self.error_handler.fatal("{0} does not belong to model {1}: {2!s}".format(typename, self.name, mobj))

    def unsupported_operator_error(self, left_arg, op, right_arg):
        self.fatal("Unsupported operation {0!s} {1!s} {2!s}, only <=, ==, >= are allowed", left_arg, op, right_arg)

    def _check_cplex_version(self, verbose=True):
        self_env = self.environment
        ok = True
        if self_env.has_cplex:
            env_cplex_version = self_env.cplex_version
            assert env_cplex_version
            if env_cplex_version < self.parameters.cplex_version:
                if verbose:
                    self.error(
                        "CPLEX DLL has revision: {0}, older than parameter version: {1}. Parameters are disabled.",
                        env_cplex_version, self.parameters.cplex_version)
                ok = False
        return ok

    def round_nearest(self, x):
        # INTERNAL
        return round_nearest_towards_infinity(x, self.infinity)

    def __init__(self, name='docplex_model',
                 output_level=InfoLevel.INFO,
                 solver_code='default',
                 docloud_context=None):
        """__init__(name='docplex_model', output_level=InfoLevel.INFO, docloud_context=None)
        """
        self._name = name

        self.__solver_code = solver_code
        self.__error_handler = DefaultErrorHandler(output_level)

        # automatic name patterns
        self._automatic_varname_pattern = "_x"
        # noinspection PyPep8
        self._automatic_ctname_pattern = "_c"
        self._automatic_indname_pattern = "_ic"

        #
        self.__allvarctns = []
        self.__single_vars = []
        self.__allvars = []
        self.__vars_by_name = {}
        self.__vars_by_index = None
        self.__allcts = []
        self.__cts_by_name = {}
        self.__cts_by_index = None
        self._kpis_by_name = OrderedDict()
        self._progress_listeners = []
        self._solve_hooks = []  # debugSolveHook()
        self.__mip_start = None

        # -- float formats
        self._float_precision = 4
        self._var_format_map = {VarType.Binary: "%d",
                                VarType.Integer: "%d",
                                VarType.Continuous: "%.4f"}

        # -- scopes
        self._var_scope = _SymbolGenerator(self.automatic_var_name_pattern)
        self._ct_scope = _SymbolGenerator(self.automatic_ct_name_pattern)
        self._indicator_scope = _SymbolGenerator(self.automatic_indicator_name_pattern)
        self._scopes = [self._var_scope, self._ct_scope, self._indicator_scope]

        # init engine
        engine = self._make_new_engine(self.solver_code, docloud_context)
        self.__engine = engine
        self._docloud_context = docloud_context
        self.__factory = ModelFactory(self, engine, self.__error_handler)
        self._solve_count = 0
        self.__solution = None

        # all the following must be placed after an engine has been set.
        self.__factory.init()
        self._default_objective_expr = self.linear_expr(constant=1)
        self.__objective_expr = self._default_objective_expr
        self.__objective_sense = ObjectiveSense.default_sense()

        # this flag controls the column-wise formatting of text exports.
        self._format_export = True

        # parameters
        self.parameters = root_params_make(self)
        self_cplex_parameters_version = self.parameters.cplex_version

        if self.environment.has_cplex:
            if self.environment.cplex_version > self_cplex_parameters_version:
                # cplex is more recent than parameters. must update defaults.
                self.trace("* reset parameter defaults, cplex version is: {0}"  # pragma : no cover
                           .format(self_cplex_parameters_version))  # pragma : no cover
                nb = self.parameters.resync_defaults()
                if nb:
                    self.trace("#parameter defaults have been reset: %d" % nb)

    @property
    def infinity(self):
        """ Returns the numerical value used as upper bound of continuous variables.

        Note
            CPLEX usually sets this limit to 1e+20.
        """
        return self.__engine.getInfinity()

    def get_name(self):
        return self._name

    def _set_name(self, name):
        self._check_name(name)
        self._name = name

    def _check_name(self, new_name):
        if not is_string(new_name):
            self.fatal("model.name requires a valid string, got: {0!s}", new_name)
        elif not new_name:
            self.fatal("model.name requires a non-empty string, got: {0!s}", new_name)
        elif new_name.find(" ") >= 0:
            self.warning("Model name contains whitespaces: |{0:s}|", new_name)
        else:
            pass

    name = property(get_name, _set_name)

    def _get_format_export(self):
        """
        This property is used to enable or disable the formatting of
        the exported LP model.
        By formatting, we mean the limitation to 80 columns of the generated LP.
        """
        return self._format_export

    def _set_format_export(self, do_format):
        self._format_export = bool(do_format)

    format_export = property(_get_format_export, _set_format_export)

    # you can change the way variables are named by default, just define a string here.
    def get_automatic_var_name_pattern(self):
        return self._automatic_varname_pattern

    def set_automatic_var_name_pattern(self, varname):
        self._automatic_varname_pattern = varname

    automatic_var_name_pattern = property(get_automatic_var_name_pattern, set_automatic_var_name_pattern)

    def get_automatic_ct_name_pattern(self):
        return self._automatic_ctname_pattern

    def set_automatic_ct_name_pattern(self, ctname):
        self._automatic_ctname_pattern = ctname

    automatic_ct_name_pattern = property(get_automatic_ct_name_pattern, set_automatic_ct_name_pattern)

    def get_automatic_indicator_name_pattern(self):
        return self._automatic_indname_pattern

    def set_automatic_indicator_name_pattern(self, indname):
        self._automatic_indname_pattern = indname

    automatic_indicator_name_pattern = property(get_automatic_indicator_name_pattern,
                                                set_automatic_indicator_name_pattern)

    def _create_automatic_varname(self):
        return self._var_scope.new_symbol()

    def _create_automatic_ctname(self, ct):
        # need to find the proper scope: linct or indicators
        return self._get_ct_scope(ct).new_symbol()

    def get_float_precision(self):
        return self._float_precision

    def set_float_precision(self, nb_digits):
        used_digits = nb_digits
        if nb_digits < 0:
            self.warning("Negative float precision given: {0}, using 0 instead", nb_digits)
            used_digits = 0
        else:
            max_digits = self.environment.max_nb_digits
            bitness = self.environment.bitness
            if nb_digits > max_digits:
                self.warning("Given precision of {0:d} goes beyond {1:d}-bit capability, using maximum: {2:d}".
                             format(nb_digits, bitness, max_digits))
                used_digits = max_digits
        self._float_precision = used_digits
        # recompute float format
        self._var_format_map[VarType.Continuous] = "%%.%df" % nb_digits

    float_precision = property(get_float_precision, set_float_precision)

    @property
    def time_limit_parameter(self):
        """  Returns the time limit parameter.

        """
        return self.parameters.timelimit

    def get_time_limit(self):
        """ Returns the time limit for the model.

        """
        return self.time_limit_parameter.get()

    def set_time_limit(self, time_limit):
        """ Set a time limit for solve operations.

        Time limit must be a positive number.

        Args:
            time_limit: The new time limit.

        """
        self.typecheck_num(time_limit)
        if time_limit < 0:
            self.fatal("Negative time limit: {0}", time_limit)
        elif time_limit <= 1:
            self.warning("Time limit too small: {0} - using 1 instead", time_limit)
            time_limit = 1
        else:
            pass

        self.set_parameter(self.time_limit_parameter, time_limit)

    def remove_time_limit(self):
        """  Removes the time limit.
        """
        self.reset_parameter(self.time_limit_parameter)

    @property
    def continuous_var_format(self):
        return self._var_format_map[VarType.Continuous]

    @property
    def solver_code(self):
        return self.__solver_code

    @property
    def error_handler(self):
        return self.__error_handler

    @property
    def docloud_context(self):
        """ Returns the DOcloud context attached to the model, possibly None.
        """
        return self._docloud_context

    @property
    def solution(self):
        """ Returns the current solution of the model.

        Returns None if the model has not yet been solved or if the last solved has failed.
        """
        return self.__solution

    def _get_solution(self):
        # INTERNAL
        return self.__solution

    @property
    def mip_start(self):
        """ Get the MIP start solution attached to the model.

        Returns None if no MIP start was defined.

        :return: An instance of SolveSolution if a MIP start has been defined, else None.
        """
        return self.__mip_start

    def fatal(self, msg, *args):
        self.__error_handler.fatal(msg, args)

    def error(self, msg, *args):
        self.__error_handler.error(msg, args)

    def warning(self, msg, *args):
        self.__error_handler.warning(msg, args)

    def info(self, msg, *args):
        self.__error_handler.info(msg, args)

    def trace(self, msg, *args):
        self.error_handler.trace(msg, args)

    def get_output_level(self):
        return self.__error_handler.outputLevel

    def set_output_level(self, new_output_level):
        self.__error_handler.set_output_level(new_output_level)

    def set_quiet(self):
        self.error_handler.set_quiet()

    output_level = property(get_output_level, set_output_level)

    def set_trace_mode(self, trace_mode):
        self.error_handler.set_trace_mode(trace_mode)
        self.__engine.notify_trace_mode(trace_mode)

    def enable_trace(self):
        self.set_trace_mode(True)

    def disable_trace(self):
        self.set_trace_mode(False)

    def is_trace_enabled(self):
        return self.error_handler.is_trace_enabled()

    def clear(self):
        """ Clears the model of all modeling objects.
        """
        self._clear_internal()

    def _clear_internal(self):
        self.__allvars = []
        self.__allvarctns = []
        self.__single_vars = []
        self.__vars_by_name = {}
        self.__allcts = []
        self.__cts_by_name = {}
        self._kpis_by_name = OrderedDict()  # kpis must be ordered
        self._solve_count = 0
        self._last_solve_status = JobSolveStatus.UNKNOWN  # initial status
        self.__solution = None
        self.__mip_start = None
        self._clear_scopes()

    def _clear_scopes(self):
        for a_scope in self._scopes:
            a_scope.reset()

    def _make_new_engine(self, solver_code, docloud_context):
        new_engine = EngineFactory.new_engine(solver_code, self.environment, self, docloud_context)
        return new_engine

    def _set_engine(self, e2):
        self.__engine = e2
        self.__factory.update_engine(e2)

    def _clear_engine(self, restart):
        # INTERNAL
        old_engine = self.__engine
        if old_engine:
            # dispose of old engine.
            old_engine.end()
            # from Ryan InstanceCounter
            del old_engine
            self.__engine = None
        if restart:
            new_engine = self._make_new_engine(self.solver_code, self._docloud_context)
            self._set_engine(new_engine)

    def get_engine(self):
        # INTERNAL for testing
        return self.__engine

    def ensure_setup(self):
        # defined for compatibility with AbstractModel
        pass

    def print_information(self):
        """ Prints general informational statistics on the model.

        Prints the number of variables and their breakdown by type.
        Prints the number of constraints and their breakdown by type.

        """
        if self.error_handler.prints_info():
            print("Model: %s" % self.name)
            self.get_statistics().print_information()

            self_params = self.parameters
            if self_params.has_nondefaults():
                print(" - parameters:")
                self_params.print_information(indent_level=5)  # 3 for " - " + 2 = 5
            else:
                print(" - parameters: defaults")

    def _is_empty(self):
        # INTERNAL
        return 0 == self.number_of_constraints and 0 == self.number_of_variables

    # @profile
    def __notify_new_model_object(self, descr, mobj, mindex,
                                  name_dir, idx_dir,
                                  is_name_safe=False):
        """
        Notifies the return af an object being create on the engine.
        :param descr: A string describing the type of the object being created (e.g. Constraint, Variable).
        :param mobj:  The newly created modeling object.
        :param mindex: The index as returned by the engine.
        :param name_dir: The directory of objects by name (e.g. name -> constraitn directory).
        :param idx_dir:  The directory from idices to objects.
        """
        mobj.set_index(mindex)

        if name_dir is not None:
            mobj_name = mobj._get_name()
            if mobj_name:
                # in some cases, names are checked before register
                if not is_name_safe:
                    if mobj_name in name_dir:
                        old_name_value = name_dir[mobj_name]
                        # Duplicate constraint name: foo
                        self.fatal("Duplicate {0} name: {1} already used for {2!s}", descr, mobj_name, old_name_value)

                name_dir[mobj_name] = mobj

        # store in idx dir if any
        if idx_dir is not None:  # do not use if idx_dir because an empty dir is False
            idx_dir[mindex] = mobj

    def _register_one_var(self, var, var_index):
        self.__notify_new_var(var, var_index)
        self._var_scope.notify_new_index(var_index)
        #
        self.__allvars.append(var)

    # @profile
    def _register_block_vars(self, allvars, indices):
        # with MyTimer("register_vars", verbose=True) as t:
        for k in fast_range(len(allvars)):
            self.__notify_new_var(allvars[k], indices[k])
        # increment global sequence container
        self.__allvars.extend(allvars)
        max_var_index = indices[-1]  # max index is last index
        self._var_scope.notify_new_index(max_var_index)

    def __notify_new_var(self, var, var_index):
        self.__notify_new_model_object("variable", var, var_index,
                                       self.__vars_by_name, self.__vars_by_index)

    def _get_ct_scope(self, ct):
        return self._indicator_scope if isinstance(ct, IndicatorConstraint) else self._ct_scope

    def _register_one_constraint(self, ct, ct_index, is_ctname_safe=False):
        """
        INTERNAL
        :param ct: The new constraint to register.
        :param ct_index: The index as returned by the engine.
        :param is_ctname_safe: True if ct name has been checked for duplicates already.
        :return:
        """
        self._get_ct_scope(ct).notify_new_index(ct_index)

        self.__notify_new_model_object(
            "constraint", ct, ct_index,
            self.__cts_by_name, self.__cts_by_index,
            is_name_safe=is_ctname_safe)

        self.__allcts.append(ct)

    def _register_block_cts(self, cts, indices, safe_names=False):
        nb_cts = len(cts)
        max_ct_index = -1
        for c in fast_range(nb_cts):
            ct_idx = indices[c]
            self.__notify_new_model_object("constraint", cts[c], ct_idx,
                                           self.__cts_by_name, self.__cts_by_index,
                                           is_name_safe=safe_names)
            if ct_idx > max_ct_index:
                max_ct_index = ct_idx
        self._ct_scope.notify_new_index(max_ct_index)
        # extend container faster than append()
        self.__allcts.extend(cts)

    # iterators
    def iter_var_containers(self):
        # INTERNAL
        return iter(self.__allvarctns)

    def _iter_single_vars(self):
        # INTERNAL
        return iter(self.__single_vars)

    def _add_var_container(self, ctn):
        # INTERNAL
        self.__allvarctns.append(ctn)

    def _count_variables_with_type(self, vartype):
        type_filter = lambda dv: dv.has_type(vartype)
        return sum(1 for _ in filter(type_filter, self.__allvars))

    def iter_variables_with_type(self, vartype):
        for v in self.iter_variables():
            if v.has_type(vartype):
                yield v

    @property
    def number_of_variables(self):
        """ Returns the total number of decision variables, all types combined.

        """
        return len(self.__allvars)

    @property
    def number_of_binary_variables(self):
        """ Returns the total number of binary variables added to the model.
        """
        return self._count_variables_with_type(VarType.Binary)

    @property
    def number_of_integer_variables(self):
        """ Returns the total number of integer variables added to the model.
        """
        return self._count_variables_with_type(VarType.Integer)

    @property
    def number_of_continuous_variables(self):
        """ Returns the total number of continuous variables added to the model.
        """
        return self._count_variables_with_type(VarType.Continuous)

    def has_integer_or_binary_vars(self):
        for dv in self.iter_variables():
            if dv.is_discrete():
                return True
        return False

    def get_statistics(self):
        """ Returns statistics on the model.

        Returns a new instance of ModelStatistics.

        See Also:
            :class:`ModelStatistics`,

        """
        return ModelStatistics._make_new_stats(self)

    def iter_variables(self):
        """ Iterates over all the variables in the model.

        Returns the variables in the order they were added to the model,
        regardless of their type.

        Returns:
            An iterator object.
        """
        return iter(self.__allvars)

    def iter_binary_vars(self):
        """ Iterates over all binary variables in the model.

        Returns the variables in the order they were added to the model.

        Returns:
            An iterator object.
        """
        return self.iter_variables_with_type(VarType.Binary)

    def iter_integer_vars(self):
        """ Iterates over all integer variables in the model.

        Returns the variables in the order they were added to the model.

        Returns:
            An iterator object.
        """
        return self.iter_variables_with_type(VarType.Integer)

    def iter_continuous_vars(self):
        """ Iterates over all continuous variables in the model.

        Returns the variables in the order they were added to the model.

        Returns:
            An iterator object.
        """
        return self.iter_variables_with_type(VarType.Continuous)

    def get_var_by_name(self, name):
        """ Looks a for a variable from a name.

        Returns either a variable if it finds one with exactly this name, or None.

        Args:
            name: The name of the variable we are looking for.

        Returns:
            A variable (instance of Var) or None.
        """
        return self.__vars_by_name.get(name, None)

    # @staticmethod
    # def _build_name_dict(mobj_seq):
    # """
    # INTERNAL: builds a dictionary of names to objects, given a sequence
    # Objects with no name (that is name is None) are omitted
    #     :param mobj_seq: the sequence of objects (assumed to have a "name" attribute
    #     :return: a dicitonary of names to objects
    #     """
    #     return {mobj.name: mobj for mobj in mobj_seq if mobj.name is not None}

    # index management
    def _build_index_dict(self, mobj_seq, raise_on_invalid_index=False):
        """
        Lazily creates a dict from indices to objects (variable or constraints)
        :param mobj_seq:
        :param raise_on_invalid_index:
        :return:
        """
        idx_dict = {}
        for mobj in mobj_seq:
            if mobj.has_valid_index():
                idx_dict[mobj.index] = mobj
            elif raise_on_invalid_index:  # pragma : no cover
                self.fatal("Object has invalid index: {0!s}", mobj)  # pragma : no cover
        return idx_dict

    def get_var_by_index(self, idx):
        # INTERNAL
        self.typecheck_valid_index(idx)
        if not self.__vars_by_index:
            self.__vars_by_index = self._build_index_dict(self.__allvars)
        return self.__vars_by_index.get(idx)

    def set_var_ub(self, var, new_ub):
        # INTERNAL: use var.ub property to set ub
        if new_ub is not None:
            self.typecheck_num(new_ub, "set_var_ub")
            if new_ub != var.ub:
                self.__engine.set_var_attribute(var, "ub", new_ub)
                var._internal_set_ub(new_ub)

    def set_var_lb(self, var, new_lb):
        # INTERNAL: use var.lb to set lb
        if new_lb is not None:
            self.typecheck_num(new_lb, "set_var_ub")
            if new_lb != var.lb:
                self.__engine.set_var_attribute(var, "lb", new_lb)
                var._internal_set_lb(new_lb)

    def _set_var_type(self, d, new_vartype):
        # DOES NOT WORK: CPLEX does not allow this..
        if self.typecheck_vartype(new_vartype):  # and new_vartype != var.vartype:
            self.__engine.setVarType(d, new_vartype)
            d._internal_set_type(new_vartype)

    def set_var_name(self, dvar, new_arg_name):
        # INTERNAL: use var.name to set variable names
        new_name = str(new_arg_name)
        if not new_name:
            self.fatal("Not a valid name: {0!s}", new_arg_name)
        elif new_name != dvar.name:
            self.__engine.set_var_attribute(dvar, "name", new_name)
            ModelingObjectBase.set_name(dvar, new_name)
        return self

    def _typecheck_bounds(self, header, candidate_bounds):
        check_ok = True
        if is_number(candidate_bounds):
            pass
        elif not is_iterable(candidate_bounds):
            check_ok = False
        else:
            check_ok = all(is_number(b) for b in candidate_bounds)
        if not check_ok:
            self.fatal("Model.{0}: expecting either number or number_sequence as bounds, got: {1!s}",
                       header, candidate_bounds)

    def set_var_lbs(self, changed_vars, new_lbs):
        """ Change lower bounds of a collection of variables.

        new_lbs accepts either a single number (all lower bouds are set to this number)
        or an array of numbers. With an array, variables are set the corresponding bound in the array.
        If the bounds array is smaller than the variable array, the remaining variables are unchanged.


        Args:
            changed_vars: The collection of variables to be changed
            new_lbs: Either a number or a collection of numbers.

        """
        self.typecheck_iterable(changed_vars)
        self._typecheck_bounds("set_var_lbs", new_lbs)

        if is_iterable(new_lbs):
            engine_new_lbs = new_lbs
        else:
            engine_new_lbs = [new_lbs] * len(changed_vars)
        actual_lbs = self.__engine.set_var_attribute(changed_vars, "lb", engine_new_lbs)
        for var, new_lb in zip(changed_vars, actual_lbs):
            var._internal_set_lb(new_lb)

    def set_var_ubs(self, changed_vars, new_ubs):
        """ Change upper bounds of a collection of variables.

        new_ubs accepts either a single number (all upper bouds are set to this number)
        or an array of numbers. With an array, variables are set the corresponding bound in the array.
        If the bounds array is smaller than the variable array, the remaining variable bounds are unchanged.

        Args:
            changed_vars: The collection of variables to be changed.
            new_ubs: Either a number or a collection of numbers.

        """
        self.typecheck_iterable(changed_vars)
        self._typecheck_bounds("set_var_ubs", new_ubs)

        new_lbs = [b for b in new_ubs] if is_iterable(new_ubs) else [new_ubs] * len(changed_vars)
        actual_ubs = self.__engine.set_var_attribute(changed_vars, "ub", new_lbs)
        for var, new_ub in zip(changed_vars, actual_ubs):
            var._internal_set_ub(new_ub)

    def get_constraint_by_name(self, name):
        """ Looks a for a constraint from a name.

        Returns either a constraint if it finds a constraint with exactly this name, or None
        if no constraint has this name.

        Will not raise an exception if not found

        Args:
            name: The name of the constraint we are looking for.

        Returns:
            A constraint or None.
        """
        return self.__cts_by_name.get(name)

    def get_constraint_by_index(self, idx):
        # INTERNAL
        self.typecheck_valid_index(idx)
        if not self.__cts_by_index:
            self.__cts_by_index = self._build_index_dict(self.__allcts)
        return self.__cts_by_index.get(idx)

    @property
    def number_of_constraints(self):
        """ Returns the total number of constraints that were added to the model.

        The number includes linear constraints, range constraints, and indicator constraints.
        """
        return len(self.__allcts)

    def iter_constraints(self):
        """ Return an iterator over all constraints in the model.

        The iterator iterates over all constraints (linear, ranges, indicators).

        """
        return iter(self.__allcts)

    def _count_constraints_with_type(self, cttype):
        type_filter = lambda ct: isinstance(ct, cttype)
        return sum(1 for _ in filter(type_filter, self.__allcts))

    def gen_constraints_with_type(self, cttype):
        for ct in self.__allcts:
            if isinstance(ct, cttype):
                yield ct

    @property
    def number_of_range_constraints(self):
        """ Returns the total number of range constraints added to the model.

        """
        return self._count_constraints_with_type(RangeConstraint)

    @property
    def number_of_linear_constraints(self):
        """ Returns the total number of linear constraints added to the model.

        This counts binary linear constraints (<=, >=, or ==) and
        range constraints.

        See Also:
            :func:`number_of_range_constraints`

        """
        return self._count_constraints_with_type(AbstractLinearConstraint)

    def iter_range_constraints(self):
        return self.gen_constraints_with_type(RangeConstraint)

    def iter_binary_constraints(self):
        return self.gen_constraints_with_type(BinaryLinearConstraint)

    def iter_linear_constraints(self):
        """
        Returns an iterator on the linear constraints of the model.
        This includes binary linear constraints and ranges, but not indicators.

        :return: An iterator.
        """
        return self.gen_constraints_with_type(AbstractLinearConstraint)

    def iter_indicator_constraints(self):
        """ Returns an iterator on idicator constraints in the model.

        Returns:
            An iterator object.
        """
        return self.gen_constraints_with_type(IndicatorConstraint)

    @property
    def number_of_indicator_constraints(self):
        """ Returns the number of indicator constraints in the model.
        """
        return self._count_constraints_with_type(IndicatorConstraint)

    def var(self, vartype, lb=None, ub=None, name=None):
        """ Creates a decision variable and stores it in the model.

        Args:
            vartype: The type of the decision variable;
                this can take three values: Binary, Integer or Continuous.

            lb: The lower bound of the variable; either a number or None, to use the default.
                 The default lower bound for all three variable types is zero.

            ub: The upper bound of the variable domain; expectes either a number or None to use the type's default.
                The default upper bound for Binary is 1, otherwise positive infinity.

            name: An optional string to name the variable.

        Returns:
            The newly created decision variable, an instance of Var.

        See Also:
            :class:`docplex.mp.linear.Var`,
            :func:`infinity`
        """
        self.typecheck_vartype(vartype)
        new_var = self.__factory.var(vartype, lb, ub, name)
        self.__single_vars.append(new_var)
        return new_var

    def _var(self, vartype, lb=None, ub=None, name=None):
        # INTERNAL
        new_var = self.__factory.var(vartype, lb, ub, name)
        self.__single_vars.append(new_var)
        return new_var

    def continuous_var(self, lb=None, ub=None, name=None):
        """ Creates a new continuous decision variable and stores it in the model.

        Args:
            lb: The lower bound of the variable, or None. The default is 0.
            ub: The upper bound of the variable, or None, to use the default. The default is model infinity.
            name: An optional name for the variable.

        Returns:
            An instance of the Var class with type VarType.Continuous.
        """
        return self.var(VarType.Continuous, lb, ub, name)

    def integer_var(self, lb=None, ub=None, name=None):
        """ Creates a new integer variable and stores it in the model.

        Args:
            lb: The lower bound of the variable, or None. The default is 0.
            ub: The upper bound of the variable, or None, to use the default. The default is model infinity.
            name: An optional name for the variable

        Returns:
            An instance of the Var class with type VarType.Integer.
        """
        return self.var(VarType.Integer, lb, ub, name)

    def binary_var(self, name=None):
        """ Creates a new binary variable and stores it in the model.

        Args:
            name:An optional name for the variable.

        Returns:
            An instance of the Var class with type VarType.Binary.
        """
        return self.var(VarType.Binary, name=name)

    def var_list(self, keys, vartype, lb=None, ub=None, name=str, key_format=None):
        self.typecheck_vartype(vartype)
        fixed_keys = self._make_key_seq(keys)
        ctn = _VariableContainer(self, vartype, [fixed_keys], lb, ub, name)
        self._add_var_container(ctn)
        return self.__factory.var_list(fixed_keys, vartype, lb, ub, name, 1, key_format)

    def binary_var_list(self, keys, name=str, key_format=None):
        """binary_var_list(self, keys, name=str, key_format=None)

        Create a list of binary decision variables.

        Create a list of binary variables and store them in the model.

        Args:
            keys: Either a sequence of objects or an integer.
            name: Used to name variables. Accepts either a string or
                a function. If given a string, the variable name is formed by appending the string
                to the string representation of the key object (if keys is a sequence) or the
                index of the variable within the range, if an integer argument is passed.
            key_format: A format string or None. This format string describes how keys contribute to variable names.
                        The default is "_%s". For example if name is "x" and each key object is represented by a string
                        like "k1", k2", ... then variables will be named "x_k1", "x_k2",...
                        
        Example:
            If you want each key string to be surrounded by {}, use a special key_format: "_{%s}",
            the %s denotes where the key string will be formatted and appended to "name".

        Returns:
            A list of Var objects with type VarType.Binary.

        Example:
            mdl.binary_var_list(3, "z") returns a list of size 3
            containing three binary variables with name z_0, z_1, z_2.

        """
        return self.var_list(keys, VarType.Binary, name=name, key_format=key_format)

    def integer_var_list(self, keys, lb=None, ub=None, name=str, key_format=None):
        """integer_var_list(self, keys, lb=None, ub=None, name=str, key_format=None)

        Create a list of integer decision variables.

        Create a list of variables with type VarType.Integer and store them in the model,
        and return the list.

        Args:
            keys: Either a sequence of objects or a positive integer. If passed an integer,
                it is interpreted as the number of variables to create.
            lb: Lower bounds of the variables. Accepts either a float number,
                a function (which will be called on each key argument) or None.
            ub: Upper bounds of the variables.  Accepts either a float number,
                a function (which will  be called on each key argument), or None.
            name: Used to name variables. Accepts either a string or
                a function. If given a string, the variable name is formed by appending the string
                to the string representation of the key object (if keys is a sequence) or the
                index of the variable within the range, if an integer argument is passed.
            key_format: A format string or None. This format string describes how keys contribute to variable names.
                The default is "_%s". For example if name is "x" and each key object is represented by a string
                like "k1", k2", ... then variables will be named "x_k1", "x_k2",...

        Note:
            Using None as the lower bound means the default lower bound (0) is used.
            Using None as the upper bound means the default upper bound (the model's positive infinity)
            is used.

        Returns:
            A list of Var objects with type VarType.Integer.

        """
        return self.var_list(keys, VarType.Integer, lb, ub, name, key_format)

    def continuous_var_list(self, keys, lb=None, ub=None, name=str, key_format=None):
        """continuous_var_list(self, keys, lb=None, ub=None, name=str, key_format=None)

        Creates a list of continuous decision variables.

        Create a list of variables with type VarType.Integer, store them in the model,
        and return the list.

        Args:
            keys: Either a sequence of objects or a positive integer. If passed an integer,
                it is interpreted as the number of variables to create.

            lb: Lower bounds of the variables. Accepts either a float number,
                a function,  or None. Use a number if all variables share the same lower bound,
                or use a function if lower bounds vary depending on the key, in which case,
                the function will be called on each key in keys.
                None means using the default lower bound (0) is used.

            ub: Upper bounds of the variables. Accepts either a float number,
                a function  or None. Use a number if all variables share the same upper bound,
                or use a function if upper bounds vary depending on the key, in which case,
                the function will be called on each key in keys.
                None means using the default upper bound (model infinity) is used.

            name: Used to name variables. Accepts either a string or
                a function. If given a string, the variable name is formed by appending the string
                to the string representation of the key object (if keys is a sequence) or the
                index of the variable within the range, if an integer argument is passed.
                If passed a function, this function is called on each key object to generate a name.
                The default behavior is to call str() on each key object.

            key_format: A format string or None. This format string describes how keys contribute to variable names.
                        The default is "_%s". For example if name is "x" and each key object is represented by a string
                        like "k1", k2", ... then variables will be named "x_k1", "x_k2",...
        Note:
            When keys is either an empty list or the integer 0, an empty list is returned.


        Returns:
            A list of Var objects with type VarType.Continuous.

        See Also:
            :class:`docplex.mp.linear.Var`,
            :func:`infinity`

        """
        return self.var_list(keys, VarType.Continuous, lb, ub, name, key_format)

    def continuous_var_dict(self, keys, lb=None, ub=None, name=str, key_format=None):
        """continuous_var_dict(self, keys, lb=None, ub=None, name=str, key_format=None)

        Create a dictionary of continuous variables, indexed by key objects.

        Create a dictionary that allows retrieval of variables from business
        model objects. Keys can be either a Python collection, an iterator, or a generator.

        A key can be any Python object, with the exception of None.
        Keys are used in the naming of variables.

        Note:
            If keys is empty, this method returns an empty dictionary.

        Args:
            keys: Either a sequence of objects, an iterator, or a positive integer. If passed an integer,
                it is interpreted as the number of variables to create.
            lb: Lower bounds of the variables. Accepts either a float number,
                a function,  or None. Use a number if all variables share the same lower bound,
                or use a function if lower bounds vary depending on the key, in which case,
                the function will be called on each key in keys.
                None means using the default lower bound (0) is used.
            ub: Upper bounds of the variables. Accepts either a float number,
                a function,  or None. Use a number if all variables share the same upper bound,
                or use a function if upper bounds vary depending on the key, in which case,
                the function will be called on each key in keys.
                None means using the default upper bound (model infinity) is used.

            name: Used to name variables. Accepts either a string or
                a function. If given a string, the variable name is formed by appending the string
                to the string representation of the key object (if keys is a sequence) or the
                index of the variable within the range, if an integer argument is passed.
                If passed a function, this function is called on each key object to generate a name.
                The default behavior is to call str() on each key object.

            key_format: A format string or None. This format string describes how keys contribute to variable names.
                        The default is "_%s". For example if name is "x" and each key object is represented by a string
                        like "k1", k2", ... then variables will be named "x_k1", "x_k2",...

        Returns:
            A dictionary of Var objects (with vartype VarType.Continuous) indexed by the objects in keys.

        See Also:
            :class:`docplex.mp.linear.Var`,
            :func:`infinity`
        """
        cvars = self.continuous_var_list(keys, lb, ub, name, key_format)
        return self._make_var_dict(keyseq=keys, varseq=cvars)

    def integer_var_dict(self, keys, lb=None, ub=None, name=str, key_format=None):
        """integer_var_dict(self, keys, lb=None, ub=None, name=str, key_format=None)

        Create a dictionary of integer variables, indexed by key objects.

        Create a dictionary that allows retrieval of variables from business
        model objects. Keys can be either a Python collection, an iterator, or a generator.

        A key can be any Python object, with the exception of None.
        Keys are used in the naming of variables.

        Note:
            If keys is empty, this method returns an empty dictionary.

        Args:
            keys: Either a sequence of objects, an iterator or a positive integer. If passed an integer,
                it is interpreted as the number of variables to create.

            lb: Lower bounds of the variables. Accepts either an integer number,
                a function,  or None. Use a number if all variables share the same lower bound,
                or use a function if lower bounds vary depending on the key, in which case,
                the function will be called on each key in keys.
                None means using the default lower bound (0) is used.

            ub: Upper bounds of the variables. Accepts either an integer number,
                a function,  or None. Use a number if all variables share the same upper bound,
                or use a function if upper bounds vary depending on the key, in which case,
                the function will be called on each key in keys.
                None means using the default upper bound (model infinity) is used.

            name: Used to name variables. Accepts either a string or
                a function. If given a string, the variable name is formed by appending the string
                to the string representation of the key object (if keys is a sequence) or the
                index of the variable within the range, if an integer argument is passed.
                If passed a function, this function is called on each key object to generate a name.
                The default behavior is to call str() on each key object.

            key_format: A format string or None. This format string describes how keys contribute to variable names.
                        The default is "_%s". For example if name is "x" and each key object is represented by a string
                        like "k1", k2", ... then variables will be named "x_k1", "x_k2",...

        Returns:
            A dictionary of Var objects (with vartype VarType.Integer) indexed by the objects in keys.

        See Also:
            :func:`infinity`
        """
        ivars = self.integer_var_list(keys, lb=lb, ub=ub, name=name, key_format=key_format)
        return self._make_var_dict(varseq=ivars, keyseq=keys)

    def binary_var_dict(self, keys, name=str, key_format=None):
        """binary_var_dict(self, keys, name=str, key_format=None)

        Create a dictionary of binary variables, indexed by key objects.

        Create a dictionary that allows retrieval of variables from business
        model objects. Keys can be either a Python collection, an iterator, or a generator.

        A key can be any Python object, with the exception of None.
        Keys are used in the naming of variables.

        Note:
            if keys is empty, this method returns an empty dictionary.

        Args:
            keys: Either a sequence of objects, an iterator or a positive integer. If passed an integer,
                it is interpreted as the number of variables to create.

            name: Used to name variables. Accepts either a string or
                a function. If given a string, the variable name is formed by appending the string
                to the string representation of the key object (if keys is a sequence) or the
                index of the variable within the range, if an integer argument is passed.

            key_format: A format string or None. This format string describes how keys contribute to variable names.
                        The default is "_%s". For example if name is "x" and each key object is represented by a string
                        like "k1", k2", ... then variables will be named "x_k1", "x_k2",...

        Returns:
            A dictionary of Var objects (with vartype VarType.Binary) indexed by the objects in keys.

        See Also:
            :class:`docplex.mp.linear.Var`
        """
        bvars = self.binary_var_list(keys, name, key_format=key_format)
        return self._make_var_dict(varseq=bvars, keyseq=keys)

    def var_multidict(self, vartype, seq_of_key_seqs, lb=None, ub=None, name=str, key_format=None):
        # INTERNAL
        self.typecheck_vartype(vartype)
        self.typecheck_iterable(seq_of_key_seqs)
        fixed_keys = [self._make_key_seq(ks) for ks in seq_of_key_seqs]

        ctn = _VariableContainer(self, vartype, fixed_keys, lb, ub, name)
        self._add_var_container(ctn)
        # the sequence of keysets should answer to len(no generators here)
        dimension = len(fixed_keys)
        if dimension < 1:
            self.fatal("len of key sequence must be >= 1, got: {0}", dimension)  # pragma : no cover

        # create cartesian product of keys...
        all_key_tuples = list(product(*fixed_keys))
        cube_vars = self.__factory.var_list(all_key_tuples, vartype, lb, ub, name, dimension, key_format, False)
        var_dict = self._make_var_dict(cube_vars, all_key_tuples)
        return var_dict

    def var_matrix(self, vartype, keys1, keys2, lb=None, ub=None, name=str, key_format=None):
        return self.var_multidict(vartype, [keys1, keys2], lb, ub, name, key_format)

    def binary_var_dict2(self, keys1, keys2, name=str):
        # INTERNAL, OBSOLETE, use binary_var_matrix
        return self.binary_var_matrix(keys1, keys2, name)

    def binary_var_matrix(self, keys1, keys2, name=str):
        """binary_var_matrix(self, keys1, keys2, name=str)

        Create a dictionary of binary variables, indexed by pairs of key objects.

        Create a dictionary that allows retrieval of variables from  a tuple
        of two keys, the first one from keys1, the second one from keys2.
        In short, variables are indexed by the Cartesian product of the two key sets.

        A key can be any Python object, with the exception of None.
        Keys are used in the naming of variables.

        Note:
            If any of keys1 or keys2 is empty, this method returns an empty dictionary.

        Args:
            keys1: Either a sequence of objects, an iterator, or a positive integer. If passed an integer N,
                    it is interpreted as a range from 0 to N-1.

            keys2: Same as keys1.

            name: Used to name variables. Accepts either a string or
                a function. If given a string, the variable name is formed by appending the string
                to the string representation of the key object (if keys is a sequence) or the
                index of the variable within the range, if an integer argument is passed.

            key_format: A format string or None. This format string describes how keys contribute to variable names.
                        The default is "_%s". For example if name is "x" and each key object is represented by a string
                        like "k1", k2", ... then variables will be named "x_k1", "x_k2",...

        Returns:
            A dictionary of Var objects (with vartype VarType.Binary) indexed by all couples (k1, k2)
            with k1 in keys1 and k2 in keys2

        See Also:
            :class:`docplex.mp.linear.Var`
        """
        return self.var_multidict(VarType.Binary, [keys1, keys2], 0, 1, name=name)

    def integer_var_matrix(self, keys1, keys2, lb=None, ub=None, name=str):
        """integer_var_matrix(self, keys1, keys2, lb=None, ub=None, name=str)

        Create a dictionary of integer variables, indexed by pairs of key objects.

        Create a dictionary that allows retrieval of variables from  a tuple
        of two keys, the first one from keys1, the second one from keys2.
        In short, variables are indexed by the Cartesian product of the two key sets.

        A key can be any Python object, with the exception of None.

        Arguments lb, ub and name are interpreted exactly as in Model.integer_var_dict()
        """

        return self.var_multidict(VarType.Integer, [keys1, keys2], lb, ub, name)

    def continuous_var_matrix(self, keys1, keys2, lb=None, ub=None, name=str):
        """continuous_var_matrix(self, keys1, keys2, lb=None, ub=None, name=str)

        Create a dictionary of continuous variables, indexed by pairs of key objects.

        Create a dictionary that allows retrieval of variables from a tuple
        of two keys, the first one from keys1, the second one from keys2.
        In short, variables are indexed by the Cartesian product of the two key sets.

        A key can be any Python object, with the exception of None.

        Arguments lb, ub and name are interpreted exactly as in Model.continuous_var_dict()
        """
        return self.var_multidict(VarType.Continuous, [keys1, keys2], lb, ub, name)

    def continuous_var_cube(self, keys1, keys2, keys3, lb=None, ub=None, name=str):
        """continuous_var_cube(self, keys1, keys2, keys3, lb=None, ub=None, name=str)

        Creates a dictionay of continuous variables, indexed by triplets.

        Same as continuous_var_matrix, except that variables are indexed by triplets of
        the form (k1, k2, k3) with k1 in keys, k2 in keys2, k3 in keys3.

        See Also:
            :func:`continuous_var_matrix`
        """
        return self.var_multidict(VarType.Continuous, [keys1, keys2, keys3], lb, ub, name)

    def integer_var_cube(self, keys1, keys2, keys3, lb=None, ub=None, name=str):
        return self.var_multidict(VarType.Integer, [keys1, keys2, keys3], lb, ub, name)

    def binary_var_cube(self, keys1, keys2, keys3, name=str):
        return self.var_multidict(VarType.Binary, [keys1, keys2, keys3], name=name)

    # noinspection PyMethodMayBeStatic
    def _make_var_dict(self, varseq, keyseq):
        # INTERNAL
        self.typecheck_iterable(varseq)
        self.typecheck_iterable(keyseq)
        return dict(zip(keyseq, varseq))

    def _get_zero_expr(self):
        # INTERNAL
        return self.__factory.zero_expr

    def linear_expr(self, e=0, constant=0, name=None):
        return self.__factory.linear_expr(e, constant, name)

    def monomial_expr(self, dvar, coef=1):
        # INTERNAL
        self.typecheck_var(dvar)
        return self._monomial_expr(dvar, coef)

    def _monomial_expr(self, dvar, coef):
        return MonomialExpr(self, dvar, coef)

    def abs(self, e):
        # NOT DOCUMENTED for now
        from docplex.mp.functional import AbsExpr

        expr = self._to_linear_expr(e)
        return AbsExpr(self, expr)

    _all_zeros = frozenset({0, 0.0})

    def _to_linear_expr(self, e):
        # INTERNAL

        if isinstance(e, LinearExpr):
            return e
        # elif e in self._all_zeros:
        #     return self._get_zero_expr()  # pco only
        else:
            try:
                return e.to_linear_expr()
            except AttributeError:
                # build a linear_expr
                return self.linear_expr(e)

    def scal_prod(self, dvars, coefs):
        """
        Creates a linear expression, equal to the scalar product of a list of variables and a sequence of coefficients.

        This method allows different types of input for both arguments. The variable sequence can be either a list
        or an iterator of objects that can be converted to linear expressions, that is, variables, expressions, or numbers.
        The most usual case is variables.
        The coefficients can be either a list of numbers, an iterator over numbers, or even a plain number.
        In this last case the scalar product reduces to a sum times this coefficient.


        :param dvars: A list or an iterator to variables or expressions.
        :param coefs: A list or an iterator on numbers, or even a number.

        Note: 
           If either list or iterator is empty, this method returns zero.

        :return: A linear expression or 0.
        """
        return self.__factory.scal_prod(dvars, coefs)

    def sum(self, args):
        """ Create a linear expression summing over a sequence.


        Note: This method returns 0 if the argument is an empty list or iterator.
        
        :param args: A list of objects that can be converted to linear expressions, that is, linear expressions, decision variables, or numbers.

        :return: A linear expression or 0.
        """
        return self.__factory.sum(args)

    def ge_constraint(self, e, rhs, name=None):
        return self.__factory.ge_constraint(e, rhs, name)

    def le_constraint(self, e, rhs, name=None):
        return self.__factory.le_constraint(e, rhs, name)

    def eq_constraint(self, e, rhs, name=None):
        return self.__factory.eq_constraint(e, rhs, name)

    def _post_constraint(self, ct):
        # INTERNAL
        eng = self.__engine
        if isinstance(ct, BinaryLinearConstraint):
            return eng.createBinaryLinearConstraint(ct)
        elif isinstance(ct, RangeConstraint):
            return eng.createRangeConstraint(ct)
        elif isinstance(ct, IndicatorConstraint):
            # here check whether linear ct is trivial. if yes do not send to CPLEX
            indicator = ct
            linct = indicator.linear_constraint
            if linct.is_trivial():
                is_feasible = linct._is_trivially_feasible()
                if is_feasible:
                    self.warning("Indicator constraint {0!s} has a trivially feasible constraint (no effect)",
                                 indicator)
                    return -2
                else:
                    self.warning(
                        "indicator constraint {0!s} has a trivially infeasible constraint; variable invalidated",
                        indicator)
                    indicator.invalidate()
                    return -4
            return eng.createIndicatorConstraint(ct)
        else:
            self.fatal("Expecting binary constraint, indicator or range, got: {0!s}", ct)  # pragma : no cover

    def _notify_trivial_constraint(self, ct, ctname, is_feasible):
        if ct is None:
            arg = None
        elif ctname and not ctname.startswith("_"):
            arg = ctname
        elif ct.name and not ct.name.startswith("_"):
            arg = ct.name
        else:
            arg = str(ct)
        ct_typename = ct.short_typename() if ct is not None else "constraint"
        ct_rank = self.number_of_constraints + 1
        if arg is not None:
            if is_feasible:
                self.info("Adding trivial feasible {2}: {0!s}, rank: {1}", arg, ct_rank, ct_typename)
            else:
                self.error("Adding trivially infeasible {2}: {0!s}, rank: {1}", arg, ct_rank, ct_typename)
        else:
            if is_feasible:
                self.info("Adding trivial feasible {1}, rank: {0}", ct_rank, ct_typename)
            elif ct:
                self.error("Adding trivially infeasible {1}, rank: {0}", ct_rank, ct_typename)
            else:
                self.fatal("Adding trivially infeasible {1}, rank: {0}", ct_rank, ct_typename)

    def _prepare_constraint(self, ct, ctname, do_check=True):
        # INTERNAL
        if ct is True:
            # sum([]) == 0
            self._notify_trivial_constraint(None, ctname, is_feasible=True)
            return False
        elif ct is False:
            # happens with sum([]) and constant e.g. sum([]) == 2
            ct = self.__factory.new_trivial_infeasible_ct()
            self._notify_trivial_constraint(None, ctname, is_feasible=False)
            if ctname:
                self.fatal("Adding a trivially infeasible constraint with name: {0}", ctname)
            else:
                # analogous to 0 == 1, model is sure to fail
                self.fatal("Adding trivially infeasible constraint")
        else:
            self.typecheck_constraint(ct)
            # -- watch for trivial cts e.g. linexpr(0) <= linexpr(1)
            if ct.is_trivial():
                if ct._is_trivially_feasible():
                    self._notify_trivial_constraint(ct, ctname, is_feasible=True)
                elif ct._is_trivially_infeasible():
                    self._notify_trivial_constraint(ct, ctname, is_feasible=False)

        # is this necessary if we're sure where we are sure of the calling context??
        if do_check and not ct.is_in_model(self):
            self.fatal("Constraint {0!s} does not belong to model {1}", ct, self.name)
        # --- name management ---
        if not ctname:
            if not ct.name:
                ct.name = self._create_automatic_ctname(ct)
                ct.notify_automatic_name()
        elif ctname in self.__cts_by_name:
            self.fatal("Duplicate constraint name: {0!s}, used for: {1}", ctname,
                       self.get_constraint_by_name(ctname).to_string())
        else:
            # might be an issue if both ctname and ct.name exist
            ct.name = ctname

        assert ct.has_name()
        # ---
        # at this stage we are sure the ctname is valid whatsoever.
        if do_check and ct.has_valid_index():
            self.warning("constraint has already been posted: {0!s}, index is: {1}", ct, ct.index)  # pragma : no cover
            return False  # pragma : no cover
        return True

    # @profile
    def _add_constraint_internal(self, ct, ctname, do_check=True):
        if ct is True:
            # sum([]) == 0
            ct = self.__factory.new_trivial_feasible_ct()
            self._notify_trivial_constraint(None, ctname, is_feasible=True)
        elif ct is False:
            # happens with sum([]) and constant e.g. sum([]) == 2
            ct = self.__factory.new_trivial_infeasible_ct()
            self._notify_trivial_constraint(None, ctname, is_feasible=False)
            if ctname:
                self.fatal("Adding a trivially infeasible constraint with name: {0}", ctname)
            else:
                # analogous to 0 == 1, model is sure to fail
                self.fatal("Adding trivially infeasible constraint")
        else:
            self.typecheck_constraint(ct)
            # -- watch for trivial cts e.g. linexpr(0) <= linexpr(1)
            if ct.is_trivial():
                if ct._is_trivially_feasible():
                    self._notify_trivial_constraint(ct, ctname, is_feasible=True)
                elif ct._is_trivially_infeasible():
                    self._notify_trivial_constraint(ct, ctname, is_feasible=False)

        # is this necessary if we're sure where we are sure of the calling context??
        if do_check and not ct.is_in_model(self):
            self.fatal("Constraint {0!s} does not belong to model {1}", ct, self.name)
        # --- name management ---
        if not ctname:
            if not ct.name:
                ct.name = self._create_automatic_ctname(ct)
                ct.notify_automatic_name()
        elif ctname in self.__cts_by_name:
            self.fatal("Duplicate constraint name: {0!s}, used for: {1}", ctname,
                       self.get_constraint_by_name(ctname).to_string())
        else:
            # might be an issue if both ctname and ct.name exist
            ct.name = ctname

        assert ct.has_name()
        # ---
        # at this stage we are sure the ctname is valid whatsoever.
        if do_check and ct.has_valid_index():
            self.warning("constraint has already been posted: {0!s}, index is: {1}", ct, ct.index)  # pragma : no cover
            return ct.index  # pragma : no cover

        ct_engine_index = self._post_constraint(ct)
        self._register_one_constraint(ct, ct_engine_index, is_ctname_safe=True)
        return ct

    def _remove_constraint_internal(self, ct):
        """
        No typechecking for this internal version.
        :param ct:
        :return:
        """
        ct_name = ct.name
        ct_index = ct.unchecked_index
        # remove from engine.
        self.__engine.remove_constraint(ct)
        # remove from model.
        self.__allcts.remove(ct)
        if ct_name:
            del self.__cts_by_name[ct_name]
        if self.__cts_by_index is not None:
            del self.__cts_by_index[ct_index]
        ct.notify_deleted()
        if ModelingObject.is_valid_index(ct_index):
            # resync indices from doomed up
            for ct in self.iter_linear_constraints():
                old_model_index = ct.unchecked_index
                if old_model_index > ct_index:
                    updated_index = self.__engine.get_ct_index(ct)
                    if updated_index != old_model_index:
                        ct.set_index(updated_index)
        self._sync_constraint_indices(self.iter_linear_constraints())

    def _resolve_ct(self, ct_arg, silent=False, context=None):
        verbose = not silent
        if context:
            printed_context = context + ": "
        else:
            printed_context = ""
        if isinstance(ct_arg, AbstractConstraint):
            return ct_arg
        elif is_string(ct_arg):
            ct = self.get_constraint_by_name(ct_arg)
            if ct is None and verbose:
                self.error("{0}no constraint with name: \"{1}\" - ignored", printed_context, ct_arg)
            return ct
        else:
            if verbose:
                self.error("{0}unexpected argument {1!s}, expecting string or constraint", printed_context, ct_arg)
            return None

    def remove_constraint(self, ct_arg):
        """ Removes a constraint from the model.

        Args:
            ct_arg: The constraint to remove. Accepts either a constraint object or a string.
                If passed a string, looks for a constraint with that name.

        """
        ct = self._resolve_ct(ct_arg, silent=False, context="remove_constraint")
        if ct is not None:
            self._check_in_model(ct, "constraint")
            self._remove_constraint_internal(ct)

    def add_range(self, lb, expr, ub, rng_name=None):
        """ Adds a new range constraint to the model.

        A range constraint states that a linear
        expression has to stay within an interval [lb..ub].
        Both lb and ub have to be float numbers with lb smaller than ub.
        Raises an exception if lb is greater than ub.

        The method creates a new range constraint and adds it to the model.

        Args:
            lb: A float number.
            expr: A linear expression, e.g. X+Y+Z.
            ub: A float number, which should be greater than lb.
            rng_name: An optional name for the range constraint.

        See Also:
            :class:`docplex.mp.linear.RangeConstraint`

        Returns:
            The newly created range constraint, an instance of RangeConstraint.

        """
        rng = self.range_constraint(lb, expr, ub, rng_name)
        ct = self._add_constraint_internal(rng, rng_name)
        return ct

    def add_indicator(self, binary_var, linear_ct, active_value=1, name=None):
        """ Adds a new indicator constraint to the model.

        An indicator constraint links (one-way) the value of a binary variable to
        the satisfaction of a linear constraint.
        If the binary variable equals the active value then the constraint is satisfied, but
        otherwise the constraint may or may not be satisfied.

        :param binary_var: The binary variable used to control the satisfaction of the linear constraint.
        :param linear_ct: A linear constraint (EQ, LE, GE).
        :param active_value: 0 or 1. The value used to trigger the satisfaction of the constraint. The default is 1.
        :param name: An optional name for the indicator constraint.

        :return: The newly created indicator constraint.
        """
        # noinspection PyPep8
        indicator_trivial_feasible_idx = -2
        indicator_trivial_infeasible_idx = -4
        self.typecheck_var(binary_var)
        self.typecheck_constraint(linear_ct)
        self.typecheck_zero_or_one(active_value)
        self._check_in_model(binary_var, typename="binary variable")
        self._check_in_model(linear_ct, typename="linear_constraint")

        indicator = self.indicator_constraint(binary_var=binary_var, linear_ct=linear_ct, active_value=active_value)

        if linear_ct.is_trivial():
            is_feasible = linear_ct._is_trivially_feasible()
            if is_feasible:
                self.warning("Indicator constraint {0!s} has a trivial feasible linear constraint (has no effect)",
                             indicator)
                return indicator_trivial_feasible_idx
            else:
                self.warning("indicator constraint {0!s} has a trivial infeasible linear constraint - invalidated",
                             indicator)
                indicator.invalidate()
                return indicator_trivial_infeasible_idx
        else:
            return self._add_constraint_internal(indicator, name)

    def range_constraint(self, lb, expr, ub, rng_name=None):
        """ Creates a new range constraint, but does not add it to the model.

        A range constraint states that a linear
        expression has to stay within an interval [lb..ub].
        Both lb and ub have to be floating-point numbers with lb smaller than ub.
        Raises an exception if lb is greater than ub.

        The method creates a new range constraint but does not add it to the model.

        Args:
            lb: A float number.
            expr: A linear expression, e.g. X+Y+Z.
            ub: A float number, which should be greater than lb.
            rng_name: An optional name for the range constraint.

        Returns:
            The newly created range constraint.

        """
        self.typecheck_num(lb, 'Model.range_constraint')
        self.typecheck_num(ub, 'Model.range_constraint')
        if not lb <= ub:
            self.error("infeasible range constraint, lb={0}, ub={1}, expr={2}", lb, ub, expr)

        expr1 = self._to_linear_expr(expr)
        rng = self.__factory.new_range_constraint(lb, expr1, ub, rng_name)
        return rng

    def indicator_constraint(self, binary_var, linear_ct, active_value=1):
        self.typecheck_var(binary_var)
        self.typecheck_constraint(linear_ct, AbstractLinearConstraint)
        actual_active_value = 1 if bool(active_value) else 0
        return self.__factory.new_indicator_constraint(binary_var, linear_ct, actual_active_value)

    def add_constraint(self, ct, ctname=None):
        """ Adds a new linear constraint to the model.

        :param ct: A linear constraint of the form <expr1> <op> <expr2>, where both expr1 and expr2 are
         linear expressions built from variables in the model, and <op> is a relational operator
         among <= (smaller or equal), == (equal), and >= (greater or equal).


        :param ctname: An optional string used to name the constraint.

        :return: The newly added constraint.
        """
        ct = self._add_constraint_internal(ct, ctname)
        return ct

    def _add_constraint_block(self, ct_ctname_seq, safe=False):
        # INTERNAL
        posted_cts = [ct for ct, ctname in ct_ctname_seq if self._prepare_constraint(ct, ctname, do_check=safe)]
        ct_indices = self.__engine.create_block_linear_constraints(posted_cts)
        self._register_block_cts(posted_cts, ct_indices)
        return posted_cts

    # ----------------------------------------------------
    # objective
    # ----------------------------------------------------

    def _clear_engine_objective(self):
        # INTERNAL
        self.__engine.clear_objective(self.__objective_expr)

    @property
    def objective_expr(self):
        """ Returns the current expression used as the model objective.
        """
        return self.__objective_expr

    @property
    def objective_sense(self):
        """ Returns the direction of the optimization.

        :return: An instance of ObjectiveSense, either Minimize or Maximize.
        """
        return self.__objective_sense

    def minimize(self, expr):
        """ Sets an expression as the expression to be minimized.

        The argument is converted to a linear expression. Accepted types are variable (instances of Var class),
        linear expressions (instance of LinearExpr), or numbers.

        :param expr: A linear expression or a variable.
        """
        self.set_objective(ObjectiveSense.Minimize, expr)

    def maximize(self, expr):
        """
        Sets an expression as the expression to be maximized.

        :param expr: A linear expression or a variable.
        """
        self.set_objective(ObjectiveSense.Maximize, expr)

    def is_minimize(self):
        """ Checks whether the model is a minimization model.

        Note that this returns True even if the expression to minimize is a constant.
        To check whether the model has a non-constant objective, use has_objective().
        """
        return self.__objective_sense is ObjectiveSense.Minimize

    def is_maximize(self):
        """ Checks whether the model is a maximization model.
        Note that this returns True even if the expression to maximize is a constant.
        To check whether the model has a non-constant objective, use has_objective().

        """
        return self.__objective_sense is ObjectiveSense.Maximize

    def objective_coef(self, dvar):
        """ Returns the objective coefficient of a variable.

        The objective coefficient is the coefficient of the variable in
        the Model's objective expression. If the variable is not explicitly
        mentioned in the objective, it returns 0.

        :param dvar: The decision variable to compute the objective coefficient.

        :return: A float number, the objective coefficient of the variable.
        """
        self.typecheck_var(dvar)
        return self._objective_coef(dvar)

    def _objective_coef(self, dvar):
        return self.__objective_expr[dvar]

    def remove_objective(self):
        """ Clears the current objective.

        This is equivalent to setting "minimize 1".
        Any subsequent solve will look only for a feasible solution.
        You can detect this state by calling has_objective() on the model.

        See Also:
            :func:`has_objective`

        """
        self.set_objective(ObjectiveSense.Minimize, self._default_objective_expr)

    def has_objective(self):
        """ Checks whether the model has a non-constant objective expression.

        A model with a constant objective will only search for a feasible solution when solved.
        This happens either if no objectibve has been assigned to the model,
        or if the objective has been removed with remove_objective.

        See Also:
            :func:`remove_objective`

        """
        return not self.__objective_expr.is_constant()

    def set_objective(self, sense, expr):
        """ Sets a new objective.

        Args:
            sense: Either an instance of ObjectiveSense, either Minimize or Maximize,
                or a string: "min" or "max"
            expr: Is converted to a linear expresion. Acceptable types are Var, LinearExpr, or plain number

        Note that using a number will not actually optimize, but only look for a feasible solution.

        """
        if expr is None:
            expr = self._default_objective_expr
        else:
            expr = self._to_linear_expr(expr)

        if not self.objective_expr.is_constant():
            # do not try to clear a constant list as CPLEX/wrapper will crash!
            self._clear_engine_objective()

        actual_sense = self._resolve_sense(sense)

        # minor issue: we reset coefficients even if they are identical.
        self.__engine.set_objective(actual_sense, expr)

        self.__objective_sense = actual_sense
        self.__objective_expr = expr

    def _can_solve(self):
        return self.__engine.can_solve()

    def _make_start_infodict(self):
        # INTERNAL
        infodict = {}
        stats = self.get_statistics()
        infodict.update(zip(["number_of_%s_vars" % tn for tn in ["binary", "integer", "ccontinuous"]],
                            [stats.number_of_binary_variables,
                             stats.number_of_integer_variables,
                             stats.number_of_continuous_variables]))
        infodict["number_of_constraints"] = self.number_of_constraints
        return infodict

    def _make_end_infodict(self):
        return self.solution.as_dict(keep_zeros=False) if self.solution is not None else dict()

    def solve(self, docloud_context=None):
        """ Starts a solve operation on the model.

        If passed a real DOcloud context, the solve operation will be started on DOcloud.
        If the argument is None, or does not contain valid DOcloud credentials,
        the solve operation will try to use CPLEX if CPLEX is available (it uses the Environment class
        to determine this).

        If CPLEX is available, the the solve operation will be performed using the native CPLEX.
        If CPLEX is not available but if the model has its own DOcloud credentials (set at model creation time),
        the the solve operation will be started on DOcloud, using the model's own DOcloudContex instance.

        If CPLEX is not available and the model has no valid credentials, an error is raised, as there is
        no way to perform the solve.

        Args:
            docloud_context: an optional instance of DOcloudContext to perform this particular solve.

        Returns:
            True if the solve operation produced a solution.
            The solution can be accessed by the solution property.
        """
        if not self.has_objective():
            self.info("No objective to optimize - searching for a feasible solution")

        if docloud_context is not None:
            if docloud_context.has_credentials():
                return self._solve_cloud(docloud_context)
            else:
                self.fatal("DOcloud context has no valid credentials: {0!s}", docloud_context)
        # from now on docloud_context is None
        elif self.environment.has_cplex:
            # if CPLEX is installed go for it
            return self._solve_local()
        elif self.docloud_context:
            # no context passed as argument, no Cplex installed, try model's own context
            return self._solve_cloud(self.docloud_context)
        else:
            # no way to solve.. really
            return self.fatal("CPLEX DLL not found: please provide DOcloud credentials")

    def _solve_local(self):
        """ Starts a solve operation on the local machine.

        Note:
        This method will never try to solve on DOcloud, regardless of whether the model
        has an attached DOcloud context.
        If CPLEX is not available, an error is raised.

        Returns:
            True if the solve produced a solution, else False.

        """
        self.notify_start_solve()

        self_solve_hooks = self._solve_hooks
        wk_hook = wk_env.solve_hook
        if wk_hook is not None:
            self_solve_hooks.append(wk_hook)
        # connect progress listeners (if any) iff problem is mip
        if self.has_integer_or_binary_vars() and self._progress_listeners:
            self.__engine.connect_progress_listeners(self._progress_listeners)

        # call notifyStart on progress listeners
        self._fire_start_solve_listeners()
        # notify hooks only on solve_local
        for h in self_solve_hooks:
            h.notify_start_solve(self._make_start_infodict())

        # --- solve is protected in try/except block
        has_solution = False
        reported_obj = 0
        engine_status = JobSolveStatus.UNKNOWN

        try:
            new_solution = self.__engine.solve(self)
            has_solution = new_solution is not None
            self._set_solution(new_solution)
            reported_obj = self._reported_objective_value()

            # store solve status as returned by the engine.
            engine_status = self.__engine.get_solve_status()
            self._last_solve_status = engine_status

        except DOcplexException as docpx_e:
            self._set_solution(None)
            raise docpx_e

        finally:
            # call notifyEnd on progress listeners
            self._fire_end_solve_listeners(has_solution, reported_obj)

            # call hooks
            for h in self_solve_hooks:
                h.notify_end_solve(has_solution,
                                   engine_status,
                                   reported_obj,
                                   self._make_end_infodict())

            # unplug worker hook if any
            if wk_hook:
                self_solve_hooks.remove(wk_hook)
        return has_solution

    def get_solve_status(self):
        """ Returns the solve status of the last successful solve.

        If the model has been solved successfully, returns the status stored in the
        model solution. Otherwise returns JobSolveStatus.UNKNOWN

        :Returns:
            An enumerated value of type JobSolveStatus.

        See Also:
            :class:`docloud.status.JobSolveStatus`
        """
        return self._last_solve_status

    def _solve_cloud(self, docloud_context):
        # see if we can reuse the local docloud engine if any?
        docloud_engine = DOcloudEngine(error_handler=self.error_handler, docloud_context=docloud_context)

        self.notify_start_solve()
        self._fire_start_solve_listeners()
        new_solution = docloud_engine.solve(self)
        self._set_solution(new_solution)
        # store solve status as returned by the engine.
        self._last_solve_status = docloud_engine.get_solve_status()
        ok = new_solution is not None
        objective_value = None
        if ok:
            objective_value = self.objective_value
        self._fire_end_solve_listeners(ok, objective_value)
        # cloud engine is transient here
        return ok

    def solve_cloud(self, docloud_context=None):
        """ Starts execution of the model on the cloud.

        This method accepts a context (an instance of DOcloudContext) to be used when
        solving on the cloud. If the context argument is None or invalid, then it will
        use the model's own instance of DOcloudContext, set at model creation time.

        Note:
            This method will always solve the model on the cloud, whether or not CPLEX
            is available on the local machine

        If docloud_context argument is None and the model has no context attached, an exception is raised.

        Args:
            docloud_context: An optional context to use on the cloud. If None, uses the model's DOcloudContext instance, if any.

        Returns:
            True if execution produces a solution, else False.

        """
        if not docloud_context:
            if self.docloud_context:
                if isinstance(self.__engine, DOcloudEngine):
                    return self.solve()
                else:
                    return self._solve_cloud(self.docloud_context)
            else:
                self.fatal("DOcloud context is None: cannot solve on the cloud")
        elif docloud_context.has_credentials():
            return self._solve_cloud(docloud_context)
        else:
            self.fatal("DOcloud context has no valid credentials: {0!s}", docloud_context)

    def notify_start_solve(self):
        # INTERNAL
        self._solve_count += 1

    def notify_solve_failed(self):
        pass

    def notify_solve_relaxed(self, relaxed_solution):
        # INTERNAL: used by relaxer
        self._set_solution(relaxed_solution)
        if relaxed_solution is not None:
            self.notify_start_solve()
        else:
            self.notify_solve_failed()

    def _resolve_sense(self, sense_arg):
        """
        INTERNAL
        :param sense_arg:
        :return:
        """
        if isinstance(sense_arg, ObjectiveSense):
            return sense_arg
        elif is_string(sense_arg):
            return ObjectiveSense.parse(sense_arg, self.error_handler, do_raise=True)

    def resolve_objective_sense(self, sense_arg):
        return ObjectiveSense.parse(sense_arg, self.error_handler, do_raise=True)

    def solve_lexicographic(self, goals,
                            senses=ObjectiveSense.Minimize,
                            abs_tolerance=1e-5,
                            relative_tolerance=1e-4,
                            dump_pass_files=False,
                            docloud_context=None):
        """ Performs a lexicographic solve from an ordered collection of goals.

        :param goals: An ordered collection of linear expressions.

        :param senses: Either an ordered collection of senses, or one sense, or None. The default is None,
         in which case the solve uses a Minimize sense. Each sense can be either a sense object,
         that is either ObjectiveSense.Minimize or Maximize, or a string "min" or "max".

        :param abs_tolerance: A float number (default is 1e-5) used as the absolute slack in
         intermediate constraints set to keep the previous passes' objectives.

        :param relative_tolerance: A float number (default is 1e-5) used as the relative slack in
         intermediate constraints set to keep the previous passes' objectives. The slack used is the maximum
         of the absolute slack and the relative slack times the objective of the pass.

        :param docloud_context: An instance of a context or None. if not None, will force all solves
         to be ran on the cloud.

        :return: True if all passes ran successfully.
        """
        old_objective_expr = self.__objective_expr
        old_objective_sense = self.__objective_sense
        if not goals:
            self.error("solve_lexicographic requires a non-empty list of goals")
            return False

        if not is_indexable(goals):
            self.fatal("solve_lexicographic requires an indexable collection of goals, got: {0!s}", goals)
        # --- senses ---
        if senses is None:
            senses = generate_constant(ObjectiveSense.Minimize)
        elif isinstance(senses, ObjectiveSense):
            senses = generate_constant(senses)
        elif is_string(senses):
            senses = generate_constant(senses)
        elif is_iterable(senses):
            pass
        else:
            self.fatal("solve_lexicographic expects as senses: None, min/max or iterable, got: {0!s}", senses)
        iter_senses = iter(senses)
        # --- senses ---

        pass_count = 0
        m = self
        current_status = False
        # currentObjective = -1
        nested_kpi_format = '   - %s ='
        results = []
        actual_goals = []
        tolerance_scheme = _ToleranceScheme(abs_tolerance, relative_tolerance)
        # keep extra constraints, in order to remove them at the end.
        extra_cts = []
        for gi, g in enumerate(goals):
            if isinstance(g, DecisionKPI):
                goal_expr = self._to_linear_expr(g.get_expression())
                goal_name = g.name
            elif isinstance(g, tuple):
                goal_expr = self._to_linear_expr(g[0])
                goal_name = g[1] or goal_expr.name
            else:
                goal_expr = self._to_linear_expr(g)
                goal_name = goal_expr.name or "pass%d" % (gi + 1)
            actual_goals.append((goal_name, goal_expr))

        try:
            for goal_name, goal_expr in actual_goals:
                if goal_expr.is_constant() and pass_count > 1:
                    self.warning("Constant expression in lexicographoic solve: {0!s}, skipped", goal_expr)
                    continue
                pass_count += 1
                next_sense = next(iter_senses)
                sense = self._resolve_sense(next_sense)

                self.trace("lexicographic: starting pass %d, %s: %s", pass_count, sense.action().lower(), goal_name)
                m.set_objective(sense, goal_expr)
                # print("-- current objective is: {0!s}".format(goal_expr))
                if dump_pass_files:
                    pass_basename = 'lexico_%s_pass%d' % (self.name, pass_count)
                    self.dump_as_lp(basename=pass_basename)
                current_status = m.solve(docloud_context)
                if current_status:
                    current_obj = m.objective_value
                    results.append(current_obj)
                    if m.error_handler.prints_trace():
                        self.trace("lexicographic: pass #%d ok with objective=%.4f" % (pass_count, current_obj))
                        m._report_lexicographic_goals(actual_goals, kpi_header_format=nested_kpi_format)
                    tolerance = tolerance_scheme.compute_tolerance(current_obj)
                    if sense.is_minimize():
                        pass_ct = m.add_constraint(goal_expr <= current_obj + tolerance,
                                                   '_ctle_pass_%d' % pass_count)
                    else:
                        pass_ct = m.add_constraint(goal_expr >= current_obj - tolerance,
                                                   '_ctge_pass_%d' % pass_count)

                    # print(">>>> new pass ct is {0!s}".format(pass_ct))
                    extra_cts.append(pass_ct)
                else:
                    self.error("lexicographic fails. pass #%d, stop!", pass_count)
                    break
        finally:
            #  print("-> start restoring model at end of lexicographic")
            while extra_cts:
                # using LIFO logic to avoid holes in indices.
                ct_to_remove = extra_cts.pop()
                # print("* removing constraint: name: {0}, idx: {1}".format(ct_to_remove.name, ct_to_remove.index))
                self._remove_constraint_internal(ct_to_remove)
            # restore objective whatsove
            self.set_objective(old_objective_sense, old_objective_expr)
            #  print("<- end restoring model at end of lexicographic")

        if current_status:
            self.info("lexicographic ok, #passes={0}, results={1!s}", pass_count, results)
            if m.error_handler.prints_info():
                m._report_lexicographic_goals(actual_goals, kpi_header_format=nested_kpi_format)
        else:
            self.warning("lexicographic failed at pass {0}", pass_count)
        return current_status

    def _has_solution(self):
        # INTERNAL
        return self.__solution is not None

    def _set_solution(self, new_solution, do_copy_values=True):
        """
        INTERNAL: Sets this solution as the model's current solution.
        Copies values to variables (for now, let's think more about this)
        :param new_solution:
        :return:
        """
        self.__solution = new_solution
        if new_solution and do_copy_values:
            new_solution._copy_var_values()

    def check_has_solution(self):
        # see if we can refine messages here...
        if not self._has_solution():
            self.fatal("Model<{0}> has not been solved OK", self.name)

    def set_mip_start(self, mip_start_sol):
        """  Sets a solution as the  MIP starting point.

        Use a (possibly partial) solution object as starting point for a MIP.

        This is valid only for models with binary or integer variables.

        Presently, this is also only valid for a native solve (when CPLEX DLL is installed) but not on DOcloud.

        :param mip_start_sol: The solution object to use as a starting point. This argument should be of type
         SolveSolution

        """
        if not mip_start_sol:
            self.fatal("Invalid MIP start: {0!s}", mip_start_sol)
        elif not self.has_integer_or_binary_vars():
            self.error("Problem has no Integer or Binary variable, cannot use MIP start")
            return
        else:
            if mip_start_sol.check_as_mip_start(self.error_handler):
                self.__mip_start = mip_start_sol

    # def getObjectiveValue(self):
    #     # PCO: not documented, to be replaced by objective_value property
    #     return self.objective_value

    @property
    def objective_value(self):
        """ Returns the value of the objective expression in the last solve solution.

        Raises an exception if the model has not been solved successfully.

        """
        self.check_has_solution()
        return self._objective_value()

    def _objective_value(self):
        return self.solution.objective_value

    def _reported_objective_value(self, failure_obj=0):
        return self.solution.objective_value if self.solution else failure_obj

    def _resolve_path(self, path_arg, basename_arg, extension):
        # INTERNAL
        if is_string(path_arg):
            if os.path.isdir(path_arg):
                if path_arg == ".":
                    path_arg = os.getcwd()
                return self._make_output_path(extension, basename_arg, path_arg)
            else:
                # ensure path ends with extension
                return path_arg if path_arg.endswith(extension) else path_arg + extension
        else:
            assert path_arg is None
            return self._make_output_path(extension, basename_arg, path_arg)

    def _make_output_path(self, extension, basename_arg, path=None):
        self_name = self.name
        raw_basename = resolve_pattern(basename_arg, self_name) if basename_arg else self_name
        basename = normalize(raw_basename)
        output_dir = path or tempfile.gettempdir()
        path = os.path.join(output_dir, basename + extension)
        return path

    __printer_ctor_dict = {LP_format: LPModelPrinter
                           }

    lp_format = LP_format

    def _get_printer_type(self, exchange_format, do_raise=False):
        """
        :param exchange_format: The format to be used.
        :param do_raise:  A Boolean, raise exception if format is unuspported.
        :return: printer or None.
        """
        printer = self.__printer_ctor_dict.get(exchange_format)
        if not printer:
            if do_raise:
                self.fatal("Unsupported output format: {0!s}", exchange_format)
            else:
                self.error("Unsupported output format: {0!s}", exchange_format)
        return printer

    def dump_as_lp(self, path=None, basename=None):
        return self.dump(path, basename, exchange_format=LP_format)

    def export_as_lp(self, basename=None, path=None, hide_user_names=False):
        """ Exports a model in LP format.

        Args:
            basename: Controls the basename with which the model is printed.
                Accepts None, a plainstring or a string format.
                if None, uses the model's name;
                if passed a plain string, the string is used in place of the model's name;
                if passed a string format (either with %s or {0}, it is used to format the
                model name to produce the basename of the written file.

            path: A path to write file, expects a string path or None.
                can be either a directory, in which case the basename
                that was computed with the basename argument, is appended to the directory to produce
                the file.
                If given a full path, the path is directly used to write the file, and
                the basename argument is not used.
                If passed None, the output directory will be tempdir.gettempdir().

            hide_user_names: A Boolean indicating whether or not to keep user names for
                variables and constraints. If False, all names are replaced by x1, x2,... for variables,
                and c1,c2,... for constraints.

        Examples:
            Assuming the model's name is mymodel:
                m.export_as_lp() will write "mymodel.lp" in gettempdir()
                m.export_as_lp(basename="foo") will write "foo.lp" in gettempdir()
                m.export_as_lp(basename="foo", path="e:/home/docplex")
                    will write file "e:/home/docplex/foo.lp"
                m.export_as_lp("e/home/docplex/bar.lp")
                    will write file "e:/home/docplex/bar.lp"
                m.export_as_lp(basename="docplex_%s", path="e/home/")
                    will write file "e:/home/docplex/docplex_mymodel.lp"

        See Also:
            tempdir.gettempdir()

        """
        self.export(path, basename, hide_user_names, exchange_format=LP_format)

    def dump(self, path=None, basename=None, hide_user_names=False, exchange_format=LP_format):
        return self._export(path, basename,
                            use_engine=True,
                            hide_user_names=hide_user_names,
                            exchange_format=exchange_format)

    def export(self, path=None, basename=None,
               hide_user_names=False, exchange_format=LP_format):
        # INTERNAL
        return self._export(path, basename,
                            hide_user_names=hide_user_names,
                            exchange_format=exchange_format)

    def _export(self, path=None, basename=None,
                use_engine=False, hide_user_names=False,
                exchange_format=LP_format):
        # INTERNAL
        # path is either a path string or None
        if path is not None and not is_string(path):
            self.fatal("Expecting path string or None, got: {0!s}", path)
        if basename is not None and not is_string(basename):
            self.fatal("Expecting basename string or None, got: {0!s}", basename)
        # INTERNAL
        extension = ""
        try:
            extension = exchange_format.extension
        except AttributeError:
            self.fatal("Not a supported exchange format: {0!s}", exchange_format)
        raw_basename = resolve_pattern(basename, self.name) if basename else self.name
        basename = normalize(raw_basename)

        path = self._resolve_path(path, basename, extension)
        ret = self._export_to_path(path, hide_user_names, use_engine, exchange_format)
        if ret:
            self.trace("model file: {0} overwritten", path)
        return ret

    def _export_to_path(self, path, hide_user_names=False, use_engine=False, exchange_format=LP_format):
        # INTERNAL
        self.ensure_setup()
        try:
            if use_engine:
                # rely on engine for the dump
                self.__engine.dump(path)
            else:
                # a path is not a stream but anyway it will work
                self._export_to_stream(stream=path, hide_user_names=hide_user_names, exchange_format=exchange_format)
            return path
        except IOError:
            self.error("Cannot open file: {0}, model not exported".format(path))
            return None

    def _export_to_stream(self, stream, hide_user_names=False, exchange_format=LP_format):
        printer_type = self._get_printer_type(exchange_format, do_raise=True)
        if printer_type:
            printer = printer_type()
            printer.wrap_lines = self.format_export
            if hide_user_names:
                printer.forget_user_names = True
            printer.printModel(self, stream)

    def export_to_stream(self, stream, hide_user_names=False, exchange_format=LP_format):
        """export_to_stream(self, stream, hide_user_names=False)

        Export the model to an output stream in LP format.

        A stream can be one of
            - a string, interpreted as a system path
            - None, interpreted as stdout
            - otherwise assuming a Python file-type object (e.g. a StringIO() instance)
                
        Args:
            stream: sn object defining where the output willbe sent
            
            hide_user_names: An optional Boolean indicating whether or not to keep user names for
                variables and constraints. If False, all names are replaced by x1,x2,... for variables,
                and c1,c2,... for constraints. Default is to keep user names.

        """
        self._export_to_stream(stream, hide_user_names, exchange_format)

    def export_to_string(self, hide_user_names=False, exchange_format=LP_format):
        """export_to_string(self, hide_user_names=False)

        Exports the model to a string.

        The output string contains the model in LP format.

        Args:
            hide_user_names: an optional boolean indicating whether or not to keep user names for
                variables and constraints. If False, all names are replaced by x1,x2 for variables,
                and c1,c2... for constraints. Default is to keep user names.

        Returns:
            a string, containg the model exported in LP format.
        """
        oss = StringIO()
        self._export_to_stream(oss, hide_user_names, exchange_format)
        return oss.getvalue()

    # -- supported attributes
    _supported_attributes = {"duals": _SolveAttribute("duals", False),
                             "slacks": _SolveAttribute("slacks", False),
                             "reduced_costs": _SolveAttribute("reduced_costs", True)}

    # advanced values
    def _get_engine_attribute(self, arg, clazz, attr_name):
        """
        Queries and returns some solve attribute from the last solve.
        :param arg: Object or iterable.
        :param clazz: The type of the argument.
        :param attr_name: The name of the attribute.
        :return:
        """
        attr_data = self._supported_attributes.get(attr_name)
        if not attr_data:
            self.fatal("Unsupported solve attribute: {0:s}", attr_name)

        if attr_data.requires_solved and not self._can_solve():
            self.fatal('Cannot query attribute {0}, engine has no solve capability', attr_name, self.__engine.name())

        if isinstance(arg, clazz):
            attrs = self._get_engine_attributes_internal([arg], attr_data)
            return attrs[0]
        elif is_iterable(arg):
            if not arg:
                return []
            else:
                return self._get_engine_attributes_internal(arg, attr_data)
        else:
            self.fatal("Attribute {0:s} not available for: {1:s}, expecting object or iterable", attr_name, arg)

    def _get_engine_attributes_internal(self, mobjs, attr_data):
        attr_name = attr_data.name
        is_for_vars = attr_data.is_for_vars
        if is_for_vars:
            indices = [v.index for v in self.iter_variables()]
        else:
            indices = [ct.index for ct in self.iter_constraints()]
        if not self.solution.is_attributes_fetched(attr_data.name):
            # get index to value from engine
            attr_idx_map = self.__engine.get_solve_attribute(attr_name, indices)
            self.solution._store_attribute_result(attr_name, attr_idx_map, is_for_vars)
        return self.solution.get_attribute(mobjs, attr_name)

    def dual_values(self, cts):
        self.check_has_solution()
        duals = self._get_engine_attribute(cts, AbstractConstraint, 'duals')
        return duals

    def slack_values(self, cts):
        self.check_has_solution()
        return self._get_engine_attribute(cts, AbstractConstraint, 'slacks')

    def reduced_costs(self, dvars):
        self.check_has_solution()
        return self._get_engine_attribute(dvars, Var, 'reduced_costs')

    def print_solution(self, do_filter_zeros=True):
        """  Prints the value of the model variables after a solve.

        Only valid after a successful solve. If the model has not been solved successfully, an
        exception is raised.

        :param do_filter_zeros: If True, only non-zero values are printed. Default is True.

        """
        self.check_has_solution()
        self_float_precision = self.float_precision
        printed_var_counter = 1
        for var in self.iter_variables():
            if var.has_name() and not var.is_generated():
                var_value = var.unchecked_solution_value

                if not do_filter_zeros or 0 != var_value:
                    print('{0}> \"{1}\" = {2:.{prec}f}'.format(printed_var_counter, var.name, var_value,
                                                               prec=0 if var.is_discrete() else self_float_precision))
                    printed_var_counter += 1

    def report(self):
        """  Prints the value of the objective and the KPIs.
        Only valid after a successful solve, otherwise will state the model is not solved.
        """
        if self._has_solution():
            used_prec = 0 if self.objective_expr.is_discrete() else self.get_float_precision()
            self.info("model solved with objective: {0:.{prec}f}".format(self._objective_value(), prec=used_prec))
            self.report_kpis()
        else:
            self.info("Model has not been solved successfully, no reporting done.")

    def report_kpis(self, selected_kpis=None, kpi_header_format='* KPI: %s='):
        """  Prints the values of the KPIs.
        Only valid after a successful solve.
        """
        kpi_format = kpi_header_format + self.continuous_var_format  # be safe even integer KPIs might yield floats
        printed_kpis = selected_kpis if is_iterable(selected_kpis) else self.iter_kpis()
        for kpi in printed_kpis:
            kpi_value = kpi.compute()
            print(kpi_format % (kpi.name, kpi_value))

    def _report_lexicographic_goals(self, goal_name_values, kpi_header_format):
        # INTERNAL
        kpi_format = kpi_header_format + self.continuous_var_format  # be safe even integer KPIs might yield floats
        printed_kpis = goal_name_values if is_iterable(goal_name_values) else self.iter_kpis()
        for goal_name, goal_expr in printed_kpis:
            goal_value = goal_expr.solution_value
            print(kpi_format % (goal_name, goal_value))

    def iter_kpis(self):
        """ Returns an iterator over all KPIs in the model.

        Returns:
           An iterator object.
        """
        return iter(self._kpis_by_name.values())

    def kpi_by_name(self, name, try_match=True, match_case=False, do_raise=True):
        """ Fetches a KPI from a string.

        This method fetches a KPI from a string, using either exact naming or trying
        to match a substring of the KPI name.

        Args:
            name: The string to be matched.
            try_match: A Boolean. If True, returns KPI whose name is not equal to the
                argument, but contains it. Default is True.
            match_case: If True, looks for a case-exact match, else ignores case. Default is False.
            do_raise: If True, raise an exception when no KPI is found.

        Example:
            If the KPI name is "Total CO2 Cost" then fetching with argument "co2" and match_case to False
            will succeed. If match_case if True, then no KPI will be returned.

        Returns:
            The KPI expression if found. If the search fails, either raises an exception, or returns a dummy
            constant expression with 0.
        """
        for kpi in self.iter_kpis():
            kpi_name = kpi.name
            ok = False
            if kpi_name == name:
                ok = True
            elif try_match:
                if match_case:
                    ok = kpi_name.find(name) >= 0
                else:
                    ok = kpi_name.lower().find(name.lower()) >= 0
            if ok:
                return kpi
        else:
            if do_raise:
                self.fatal("Cannot find any KPI matching: {0:s}", name)
            else:
                return self._get_zero_expr()

    def kpi_value_by_name(self, name, try_match=True, match_case=False, do_raise=True):
        """ Return a KPI value from a KPI name.

        This method fetches a KPI value from a string, using either exact naming or trying
        to match a substring of the KPI name.

        Args:
            name: The string to be matched.
            try_match: A Boolean. If True, returns KPI whose name is not equal to the
                argument, but contains it. Default is True.
            match_case: If True, looks for a case-exact match, else ignores case. Default is False.
            do_raise: If True, raise an exception when no KPI is found.

        Example:
            If the KPI name is "Total CO2 Cost" then fetching with argument "co2" and match_case to False
            will succeed. If match_case if True, then no KPI will be returned.

        Note:
            Expression KPIs require a valid solution to be computed. Make sure that the model contains
            a valid solution before computing an expression-based KPI, otherwise an exception will be raised.
            Functional KPIs do not require a valid solution.

        Returns:
            The KPI value (a float).
        """
        kpi = self.kpi_by_name(name, try_match, match_case=match_case, do_raise=do_raise)
        return kpi.compute()

    def _make_kpi(self, kpi_arg, kpi_name):
        # return a concrete instance of KPI
        from docplex.mp.kpi import DecisionKPI, FunctionalKPI

        if isinstance(kpi_arg, Expr):
            return DecisionKPI(kpi_arg, kpi_name)
        elif isinstance(kpi_arg, Var):
            return DecisionKPI(kpi_arg, kpi_name)
        elif isinstance(kpi_arg, DecisionKPI):
            return DecisionKPI(kpi_arg._dobj, kpi_arg._name)
        elif isinstance(kpi_arg, FunctionalKPI):
            other_fn_kpi = kpi_arg
            return FunctionalKPI(other_fn_kpi._function, self, other_fn_kpi.get_name())
        elif is_function(kpi_arg):
            return FunctionalKPI(kpi_arg, self, kpi_name)
        else:
            self.fatal("Cannot interpret this as a KPI: {0!s}. expecting expression, variable or function", kpi_arg)

    def add_kpi(self, kpi_arg, publish_name=None):
        new_kpi = self._make_kpi(kpi_arg, publish_name)
        new_kpi_name = new_kpi.get_name()
        if new_kpi_name in self._kpis_by_name.keys():
            self.fatal("Duplicate KPI name: {0!s}. KPI names must be unique", new_kpi_name)
        self._kpis_by_name[new_kpi_name] = new_kpi
        return new_kpi

    def _check_progress_listener(self, e):
        if not isinstance(e, ProgressListener):
            self.error("Not a progress listener: {0!s}", e)
            return False
        else:
            return True

    def add_progress_listener(self, listener):
        if self._check_progress_listener(listener):
            self._progress_listeners.append(listener)

    def remove_progress_listener(self, listener):
        if self._check_progress_listener(listener):
            self._progress_listeners.remove(listener)

    def _fire_start_solve_listeners(self):
        for l in self._progress_listeners:
            l.notify_start()

    def _fire_end_solve_listeners(self, has_solution, objective_value):
        for l in self._progress_listeners:
            l.notify_end(has_solution, objective_value)
            
    def fire_jobid(self, jobid):
        for l in self._progress_listeners:
            l.notify_jobid(jobid)
            
    def clear_progress_listeners(self):
        self._progress_listeners = []

    def prettyprint(self):
        ModelPrettyPrinter().printModel(self)

    def clone(self, new_name=None):
        """ Makes a deep copy of the model, possibly with a new name.
            Variables, constraints, and objective are copied.

        Args:
            new_name: The new name to use. If None is provided, returns a "Copy of xxx" where xxx is the original model name.

        Returns:
            An instance of Model.
        """
        return self.copy(new_name)

    def copy(self, new_name=None, new_solver_code=None):
        # INTERNAL
        copy_name = new_name or "Copy of %s" % self.name
        copy_solver_code = new_solver_code or self.solver_code
        copy_model = Model(name=copy_name, solver_code=copy_solver_code)

        # clone variable containers
        for ctn in self.iter_var_containers():
            copy_model._add_var_container(ctn.copy(copy_model))

        # clone variables
        var_mapping = {}
        for v in self.iter_variables():
            copied_var = copy_model.var(v.vartype, v.lb, v.ub, v.name)
            var_mapping[v] = copied_var

        # clone constraints
        for ct in self.iter_constraints():
            copied_ct = ct.copy(copy_model, var_mapping)
            copy_model.add_constraint(copied_ct, ct.name)

        # clone objective
        if self.has_objective():
            copy_model.set_objective(self.objective_sense, self.objective_expr.copy(copy_model, var_mapping))

        # clone kpis
        for kpi in self.iter_kpis():
            copy_model.add_kpi(kpi.copy(copy_model, var_mapping))

        if self._docloud_context:
            copy_model._docloud_context = self._docloud_context.copy()

        # clone params (some day)
        return copy_model

    def _is_fully_indexed(self):
        """
        INTERNAL: Returns true if all objects have valid index, else None.
        :return: True if all variables and constraints have a valid index, else False.
        """
        justifier = None
        for dv in self.iter_variables():
            if not dv.has_valid_index():
                justifier = dv
                break
        for ct in self.iter_constraints():
            if not ct.has_valid_index():
                justifier = ct
                break
        return justifier is None

    def refresh_model(self, do_setup=True):
        # compatibility with AbtractModel
        pass

    def setup(self):
        # compatibility with AbtractModel
        pass

    def is_free_lb(self, var_lb):
        return self.__factory.is_free_lb(var_lb)

    def is_free_ub(self, var_ub):
        return self.__factory.is_free_ub(var_ub)

    def _sync_constraint_indices(self, ct_iter):
        # INTERNAL: check only when CPLEX is present.
        self_engine = self.__engine
        if self_engine.has_cplex():
            for ct in ct_iter:
                model_index = ct.get_index()
                cpx_index = self_engine.get_ct_index(ct)
                if model_index != cpx_index:
                    self.error_handler.error("indices differ, obj: {0!s}, docplex={1}, CPLEX={2}",
                                             (ct, model_index, cpx_index))

    def _sync_var_indices(self):
        self_engine = self.__engine
        if self_engine.has_cplex():
            for dvar in self.iter_variables():
                model_index = dvar.get_index()
                cpx_index = self_engine.get_var_index(dvar)
                if model_index != cpx_index:
                    self.error_handler.error("indices differ, obj: {0!s}, docplex={1}, CPLEX={2}",
                                             (dvar, model_index, cpx_index))

    def end(self):
        """ Terminates a model instance.

        Reclaims all memory consumed by the model.
        Reclaims memory allocated by CPLEX, if any.

        """
        self.clear()
        self._clear_engine(restart=False)

    def _resolve_parameter(self, param_arg):

        if isinstance(param_arg, Parameter):
            parameter = param_arg
        elif is_int(param_arg):
            # find parameter with this cplex id
            parameter = self.parameter_from_id(param_arg)
            if not parameter:
                self.error("No parameter with id: {0} - ignored", param_arg)
        else:
            parameter = None
            self.fatal("set_parameter expected Parameter instances or id, got: {0!s}", param_arg)
        return parameter

    def get_parameter(self, param_arg):
        parameter = self._resolve_parameter(param_arg)
        return self._get_parameter_internal(parameter, check_is_verbose=True)

    def _get_parameter_internal(self, parameter, check_is_verbose=True):
        if self._check_cplex_version(verbose=check_is_verbose):
            return self.__engine.get_parameter(parameter)
        else:
            return parameter.default_value

    def parameter_from_id(self, parameter_cpx_id):
        """ Finds a Parameter from a CPLEX id code.

        Args:
            parameter_cpx_id: A cplex parameter id (positive integer, e.g. 2009 is mipgap).

        Returns:
            An instance of Parameter, if found, else None
        """
        assert parameter_cpx_id >= 0
        for p in self.parameters.generate_params():
            if p.cpx_id == parameter_cpx_id:
                return p
        else:
            return None

    def set_parameter(self, param_arg, raw_value):
        parameter = self._resolve_parameter(param_arg)
        # check the value, else raise error
        new_value = parameter.check_value(raw_value, self.error_handler, do_raise=True)
        if self._check_cplex_version() and new_value is not None:
            self.__engine.set_parameter(parameter, new_value)

    def reset_parameter(self, args):
        if not self._check_cplex_version():
            return

        if args is None:
            # None argument, reset all params
            self.parameters.reset(recursive=True)
        elif is_iterable(args):
            for a in args:
                self.reset_parameter(a)
        else:
            param = self._resolve_parameter(args)
            param.set(param.default_value)

    # with protocol
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        # terminate the model upon exiting a 'with' block.
        self.end()

    def __iadd__(self, e):
        # implements the "+=" dialect a la PulP
        if not isinstance(e, AbstractConstraint):
            self.fatal("Model += can only be used with constraints, not with: {0!s}", e)
        else:
            self.add_constraint(ct=e, ctname=None)


class AbstractModel(Model):
    def __init__(self, name, output_level=1, solver_code='default', docloud_context=None):
        Model.__init__(self, name=name,
                       output_level=output_level,
                       solver_code=solver_code,
                       docloud_context=docloud_context)

    @staticmethod
    def _check_data_args(args, expected_sizemin, do_raise=True):
        msg = None
        if not args:
            msg = "Empty data collection"
        elif len(args) < expected_sizemin:
            msg = "Missing data: expecting %d collections, got: %d" % (expected_sizemin, len(args))
        else:
            pass
        if msg:
            if do_raise:
                raise DOcplexException(msg)
            else:
                print(msg)
                return False
        else:
            return True

    def is_valid(self):
        """ Redefine this function to return False if some data is invalid.
        """
        return True

    def setup_variables(self):
        raise NotImplementedError

    def setup_constraints(self):
        raise NotImplementedError

    def setup_objective(self):
        ''' Redefine this method to set the objective.
        This is not mandatory as a model might not have any objective.
        '''
        pass  # pragma: no cover

    # noinspection PyMethodMayBeStatic
    def check(self):
        ''' Redefine this method to check the model before solve.
        '''
        pass

    def setup_data(self):
        pass

    def post_process(self):
        pass

    def setup(self):
        """ Setup the model artifacts, raise exception if data are not correct.
        """
        self.setup_data()
        if self.is_valid():
            self.setup_variables()
            self.setup_constraints()
            self.setup_objective()
        else:
            self.error("model is invalid, setup stops")

    def ensure_setup(self):
        self.error_handler.ensure(self.is_valid(), "Model is not valid")
        if self._is_empty():
            self.setup()

    def restart(self):
        """ Called to restart the model in an empty state.

        The underlying model is also restarted to a clean and empty state.
        All modeling objects previsouly defined and stored in the model are discarded.
        """
        self.clear()
        # if the superclass does not call the parent class, make sure...
        self._clear_internal()
        self._clear_engine(restart=True)

    def refresh_model(self, do_setup=True):
        ''' Clears all model elements plus sets a new engine.'''
        self.restart()
        if do_setup:
            self.ensure_setup()

    def before_solve_hook(self):
        """ This method is called just before solve inside a run.
        Redefine to get some particular behavior.
        """
        if self.error_handler.prints_info():
            self.print_information()
        if self.error_handler.prints_trace():
            self.export_as_lp()

    def export(self, path=None, basename=None, hide_user_names=False, exchange_format=LP_format):
        # INTERNAL: redefine export at this stage to ensure model is setup
        self.ensure_setup()
        return Model.export(self, path, basename, hide_user_names, exchange_format)

    def run_silent(self, docloud_context=None):
        # make sure th emodel is setup
        self.ensure_setup()
        # check data and model if necessary (ddefault is do nothing)
        self.check()
        # insert some last minute code before solve.
        self.before_solve_hook()
        # call solve_run which by default calls solve
        ok = self.solve_run(docloud_context=docloud_context)
        if ok:
            self.post_process()
        return ok

    def solve_run(self, docloud_context=None):
        return self.solve(docloud_context=docloud_context)

    def run(self, docloud_context=None):
        ok = self.run_silent(docloud_context)
        if ok:
            self.report()
        return ok
