from __future__ import print_function

from six import iteritems

from docplex.mp.utils import is_int, docplex_fatal, StringIO
from docplex.mp.utils import DOcplexRedirectStdoutContext


class ParameterGroup(object):
    """ A group of parameters.

    Note:
        This class is not meant to be instantiated by users. Models come
        with a full hierarchy of parameters with groups as nodes.

    """

    def __init__(self, name, parent_group=None):
        self._name = name
        self._parent = parent_group
        self._params = []
        self._subgroups = []
        if parent_group:
            parent_group._add_subgroup(self)

    def to_string(self):
        return "group<%s>" % self.qualified_name()

    def __str__(self):
        return self.to_string()

    @property
    def name(self):
        """ Returns the name of the group.

        Note:
            Parameter group names are always lowercase.
        """
        return self._name

    def iter_params(self):
        """ Iterates over the group's own parameters.

        This method returns an iterator over the group's parameters, non-recursively.
        """
        return iter(self._params)

    @property
    def number_of_params(self):
        """  Returns the number of parameters in the group, not including subgroups.
        """
        return len(self._params)

    def recursive_number_of_params(self):
        """  Returns the total number of parameters inside the group.

        This includes all parameters of subgroups recursively.
        """
        subparams = sum(g.recursive_number_of_params() for g in self._subgroups)
        return subparams + self.number_of_params

    def recursive_number_of_groups(self):
        subgroups = sum(g.recursive_number_of_groups() for g in self._subgroups)
        return subgroups + 1

    @property
    def number_of_subgroups(self):
        """ Returns the number of subgroups of the group, non-recursively.

        """
        return len(self._subgroups)

    def iter_subgroups(self):
        return iter(self._subgroups)

    @property
    def parent_group(self):
        """ Returns the parent group, if any of the group.

        Returns:
            An instance of ParameterGroup, or None for the root group.

        """
        return self._parent

    def _add_param(self, param):
        # internal
        self._params.append(param)

    def _add_subgroup(self, subgroup):
        # internal
        self._subgroups.append(subgroup)

    def is_root(self):
        """ Checks whether the group is the root group ,i.e., has no parent group.
        """
        return self._parent is None

    def root_group(self):
        group = self
        while not group.is_root():
            group = group.parent_group
        return group

    def qualified_name(self, sep="."):
        """ Returns a string representation of the parameter hierarchy.

        Computes a string with all the parents of the parameters.

        Args:
            sep: The separator string. Default is ".".

        Example:
            parameter mip.mip_cuts.Cover returns "mip.mip_cuts.Covers".

        Returns:
            A string.
        """
        self_parent = self._parent
        if not self_parent:
            return self.name
        elif self_parent.is_root():
            return self.name
        else:
            return "".join([self._parent.qualified_name(), sep, self.name])

    def prettyprint(self, indent=0):
        tab = indent * 4 * " "
        print("{0}{1!s}={{".format(tab, self.qualified_name()))
        for p in self.iter_params():
            print("{0}    {1!s}".format(tab, p))
        for sg in self.iter_subgroups():
            assert isinstance(sg, ParameterGroup)
            sg.prettyprint(indent + 1)
        print("{0}}}".format(tab))

    def _update_self_dict(self, extra_dict, do_check=True):
        self_dict = self.__dict__
        if do_check:
            # new entries should not already be present in self.dict
            for k in extra_dict:
                if k in self_dict:
                    print("!! update_self_dict: name collision with: %s" % k)
        self_dict.update(extra_dict)

    @staticmethod
    def make(name, param_dict_fn, subgroup_fn, parent=None):
        # INTERNAL
        # factory method to create one group from:
        # 1. a lambda function taking a group as argument and returning a dict of name: param instances
        # 2. a dict of subgroup name: subgroup_make functions
        # 3. a possibly-None parent group. If None, we are at root.
        group = ParameterGroup(name, parent) if parent else RootParameterGroup(name, cplex_version=None)
        group._initialize(param_dict_fn, subgroup_fn)
        return group

    def _initialize(self, param_dict_fn, subgroup_fn):
        param_dict = param_dict_fn(self)
        self._update_self_dict(param_dict)
        if subgroup_fn:
            subgroup_fn_dict = subgroup_fn()
            subgroup_dict = {group_name: group_fn(self)
                             for group_name, group_fn in iteritems(subgroup_fn_dict)}
            self._update_self_dict(subgroup_dict)

    def number_of_nondefaults(self, ):
        return sum(1 for _ in self.generate_nondefault_params())

    def has_nondefaults(self):
        for _ in self.generate_nondefault_params():
            return True
        else:
            return False

    def reset(self, recursive=False):
        """ Resets all parameters in the group.

        Args:
            recursive: If True, also resets the subgroups.
        """
        for p in self.iter_params():
            p.reset()
        if recursive:
            for g in self.iter_subgroups():
                g.reset(recursive=True)

    def resync_defaults(self):
        # INTERNAL
        nb_resyncs = 0
        for p in self.iter_params():
            p_value = p.get()
            if p_value != p.default_value:
                p._default_value = p_value
                nb_resyncs += 1
        return nb_resyncs

    def generate_params(self):
        """  Returns a generator traversing all parameters.

        The generator yields all parameters from the group
        and also from its subgroups, recursively.
        Called from the root parameter group, returns all parameters.

        Returns:
            A generator object.

        """
        return self._generate_and_filter_params(predicate=None)

    def _generate_and_filter_params(self, predicate):
        """ A filtering generator function that traverses a group's parameters.

        This generator function traverses the group and its subgroup tree,
        yielding only  parameters that are accepted by th epredicate.

        Args:
           predicate: A function that takes one parameter as asrgument.
           The return value of this function will be interpeted as a boolean using
           Python conversionb rules.

        Returns:
            A generator object.
        """
        for p in self.iter_params():
            if predicate is None or predicate(p):
                yield p
        # now recurse
        for sg in self.iter_subgroups():
            for nd in sg._generate_and_filter_params(predicate):
                yield nd

    def generate_nondefault_params(self):
        """ A generator function that returns all non-default parameters.

        This generator function traverses the group and its subgroup tree,
        yielding those parameters with a non-default value, one at a time.
        A parameter is non-default as soon as its value differs from the default.

        Returns:
            A generator object.
        """
        return self._generate_and_filter_params(predicate=lambda p: p.is_nondefault())

    def __setattr__(self, attr_name, value):
        if attr_name.startswith("_"):
            self.__dict__[attr_name] = value
        elif hasattr(self, attr_name):
            attr = getattr(self, attr_name)
            if isinstance(attr, Parameter):
                attr.set(value)
            else:
                self.get_model().fatal("No parameter with name {0} in {1}", attr_name, self.qualified_name())
        else:
            self.get_model().fatal("No parameter with name {0} in {1}", attr_name, self.qualified_name())

    def get_model(self):
        return self.root_group().model

    def cplex_version(self):
        return self.root_group().cplex_version


class Parameter(object):
    """ Base class for all parameters.

    This class is not meant to be instantiated, but subclassed.

    """

    # noinspection PyProtectedMember
    def __init__(self, group, short_name, cpx_name, param_key, description, default_value):
        assert isinstance(group, ParameterGroup)
        self._parent = group
        self._short_name = short_name
        self._cpx_name = cpx_name
        self._id = param_key
        self._description = description
        self._default_value = default_value

        group._add_param(self)

    @property
    def short_name(self):
        return self._short_name

    @property
    def qualified_name(self, sep='.'):
        return "%s.%s" % (self._parent.qualified_name(sep=sep), self.short_name)

    @property
    def cpx_name(self):
        return self._cpx_name

    @property
    def cpx_id(self):
        return self._id

    @property
    def description(self):
        return self._description

    @property
    def default_value(self):
        return self._default_value

    def accept_value(self, new_value):
        """ Returns True if new_value is an accepd value for the parameter.

        Args:
            new_value: The candidate value.

        Returns:
            True if acceptable, else False.

        """
        raise NotImplementedError()

    def transform_value(self, raw_value):
        return raw_value

    def check_value(self, raw_value, error_handler, do_raise=True):
        if raw_value == self.default_value:
            return raw_value
        elif not self.accept_value(raw_value):
            if do_raise:
                error_handler.fatal("Value {0!s} is invalid for parameter {1}", (raw_value, self.qualified_name))
            else:
                error_handler.error("Value {0!s} is invalid for parameter {1}", (raw_value, self.qualified_name))
            return False
        else:
            return self.transform_value(raw_value)

    def _model_not_found_error(self):
        # INTERNAL
        docplex_fatal("Parameter has no model attached: {0}", self.qualified_name)

    def set(self, new_value):
        """ Changes the value of the parameter to some value.


        Args:
            new_value: The new value for the parameter.
            model: An instance of Model.

        """
        mdl = self.get_model()
        if mdl:
            mdl.set_parameter(self, new_value)

    def get(self):
        """ Returns the current value of the parameter.

        """
        mdl = self.get_model()
        if mdl:
            return mdl.get_parameter(self)
        else:
            return self.default_value

    def reset(self):
        """ Resets the parameter value to its default.
        """
        mdl = self.get_model()
        if mdl:
            mdl.reset_parameter(self)

    def is_nondefault(self):
        """ Returns True if the current value of the parameter equals its default.

        """
        mdl = self.get_model()
        return mdl.get_parameter(self) != self.default_value if mdl else False

    def is_default(self, ):
        """  Returns True if in this model, the current value of the parameter is its default.
        """
        return self.get() == self.default_value

    @staticmethod
    def _is_in_range(arg, range_min, range_max):
        if range_min is not None and arg < range_min:
            return False
        if range_max is not None and arg > range_max:
            return False
        return True

    def to_string(self):
        return "{0}:{1:s}".format(self._short_name, self.type_name())

    def __str__(self):
        return self.to_string()

    def is_numeric(self):
        return False

    def type_name(self):
        raise NotImplementedError

    def _root_group(self):
        return self._parent.root_group()

    def get_model(self):
        mdl = self._root_group().model
        if not mdl:
            self._model_not_found_error()
        return mdl


class BoolParameter(Parameter):
    def __init__(self, group, short_name, cpx_name, param_key, description, default_value):
        Parameter.__init__(self, group, short_name, cpx_name, param_key, description, default_value)

    def transform_value(self, new_value):
        if new_value in {0, 1}:
            return new_value
        elif new_value in {"true", "false"}:
            return 1 if new_value == "true" else 0
        elif new_value in {True, False}:
            return 1 if new_value else 0
        else:
            return None

    def accept_value(self, value):
        return value in {0, 1} or value in {"true", "false"} or value in {True, False}

    def type_name(self):
        return "bool"


class StrParameter(Parameter):
    def __init__(self, group, short_name, cpx_name, param_key, description, default_value):
        Parameter.__init__(self, group, short_name, cpx_name, param_key, description, default_value)
        assert isinstance(default_value, str)

    def accept_value(self, new_value):
        return isinstance(new_value, str)

    def type_name(self):
        return "string"


class IntParameter(Parameter):
    def accept_value(self, new_value):
        return is_int(new_value) and self._is_in_range(new_value, self._min_value, self._max_value)

    def is_numeric(self):
        return True

    def min_value(self):
        return self._min_value

    def max_value(self):
        return self._max_value

    def __init__(self, group, short_name, cpx_name, param_key, description, default_value, min_value=None,
                 max_value=None):
        Parameter.__init__(self, group, short_name, cpx_name, param_key, description, default_value)
        self._min_value = min_value
        self._max_value = max_value

    def type_name(self):
        return "int"


class PositiveIntParameter(IntParameter):
    def __init__(self, group, short_name, cpx_name, param_key, description, default_value, max_value=None):
        IntParameter.__init__(self, group, short_name, cpx_name, param_key, description, default_value, min_value=0,
                              max_value=max_value)

    def type_name(self):
        return "count"


class NumParameter(Parameter):
    """ A numeric parameter can take any float value, inside a range of [min,max] values.
    """

    def __init__(self, group, short_name, cpx_name, param_key, description, default_value, min_value=None,
                 max_value=None):
        Parameter.__init__(self, group, short_name, cpx_name, param_key, description, default_value)
        self._min_value = min_value
        self._max_value = max_value

    def is_numeric(self):
        return True

    def min_value(self):
        return self._min_value

    def max_value(self):
        return self._max_value

    def accept_value(self, new_value):
        return self._is_in_range(new_value, self._min_value, self._max_value)

    def type_name(self):
        return "num"

# a dictionary of formats for each type.
_param_prm_formats = {NumParameter: "%.14f",
                      IntParameter: "%d",
                      PositiveIntParameter: "%d",
                      BoolParameter: "%d",
                      StrParameter: "\"%s\""  # need quotes
                      }


class RootParameterGroup(ParameterGroup):
    """ The root parameter group (there should be only one instance at root of tree).
    """

    def __init__(self, name, cplex_version, model=None):
        ParameterGroup.__init__(self, name)
        self._cplex_version = cplex_version
        self._model = model

    def qualified_name(self, sep="."):
        return self.name

    @property
    def cplex_version(self):
        return self._cplex_version

    def is_root(self):
        return True

    @property
    def model(self):
        return self._model

    def export_prm(self, output, overload_params=None):
        """
        Exports parameters to an output stream in PRM format.

        This method writes non-default parameters in CPLEX PRM syntax.
        In addition to non-default parameters, some parameters can be forced to
        be printed with a specific value by passing a dictionary with
        Parameter objects as keys and values as asrguments.
        These values are used in the print operation, but will not be kept,
        and the values of parameters will not be changed.
        Passing None as overload_params will disable this functionality, and
        only non-default parameters are printed.

        Args:
            output: The output stream, typycally a filename.

            overload_params: A dictionary of overloaded values for
                certain parameters. This dictionary is of the form {param: value}
                the printed PRM file will use overloaded values
                for those parameters present in the dictionary.

        """
        mdl = self.model
        with DOcplexRedirectStdoutContext(output, mdl.error_handler):
            cplex_version_string = self._cplex_version
            print("# -- This content is generated by DOcplex")
            print("CPLEX Parameter File Version %s" % cplex_version_string)

            param_generator = self.generate_params()
            for param in param_generator:
                if overload_params and param in overload_params:
                    param_value = overload_params[param]
                else:
                    param_value = mdl._get_parameter_internal(param, check_is_verbose=False)

                if param_value != param.default_value:
                    print("{0:<33}".format(param.cpx_name), end="")
                    print(_param_prm_formats[type(param)] % param_value)

            print("# --- end of generated prm data ---")

    def print_information(self, indent_level=0):
        self.write_parameter_info_to_stream(output=None, overload_params=None, indent_level=indent_level)

    def write_parameter_info_to_stream(self, output, overload_params=None, indent_level=0):
        """ Writes parameters to an output stream.

        This method writes non-default parameters in a human readable syntax.
        In addition to non-default parameters, some parameters can be forced to
        be printed with a specific value by passing a dictionary with
        Parameter objects as keys and values as arguments.
        These values are used in the print operation but will not be kept,
        and the values of parameters will not be changed.
        Passing None as overload_params will disable this functionality, and
        only non-default parameters are printed.

        Args:
            mdl: The model to which parameters belong.
            output: The output stream.
            overload_params: A dictionary of overloaded values for
                certain parameters. This dictionary is of the form {param: value}.
        """
        error_handler = self.model.error_handler
        indent = " " * indent_level
        with DOcplexRedirectStdoutContext(output, error_handler) as redirect:
            param_generator = self.generate_params()
            root_param_name = "parameters"
            for param in param_generator:
                if overload_params and param in overload_params:
                    param_value = overload_params[param]
                elif param.is_nondefault():
                    param_value = param.get()
                else:
                    param_value = None
                if param_value is not None:
                    print("{0}{1}.{2} = {3!s}"
                          .format(indent, root_param_name,
                                  param.qualified_name,
                                  _param_prm_formats[type(param)] % param_value))

    def export_prm_to_string(self, overload_params=None):
        """  Exports non-default parameters in PRM format to a string.

        The logic of overload is the same as in export_prm.
        A parameter is written if either it is a key on overload_params,
        or it has a non-default value.
        This allows merging non-default params with temporary parameter values.


        Args:
            mdl: The model to which parameters belong.
            overload_params: A dictionary of overloaded values, possibly None.

        Note:
            This method causes no side effects on the model's parameters.

        See Also:
            :func:`export_prm`

        Returns:
            A string, in CPLEX PRM format.
        """
        oss = StringIO()
        self.export_prm(oss, overload_params)
        return oss.getvalue()

    def write_parameter_info_to_string(self, overload_params=None):
        """  Writes parameters in readable format to a string.

        The logic of overload is the same as in export_prm.
        A parameter is written if either it is a key on overload_params,
        or it has a non-default value.
        This allows merging non-default params with temporary parameter values.


        Args:
            overload_params: A dictionary of overloaded values, possibly None.

        Note:
            This method causes no side effects on the model's parameters.

        See Also:
            :func:`write_parameters`

        Returns:
            A string.
        """
        oss = StringIO()
        self.write_parameter_info_to_stream(oss, overload_params)
        return oss.getvalue()

    @staticmethod
    def make(name, param_dict_fn, subgroup_fn, cplex_version, mdl):
        # INTERNAL
        # factory method to create one group from:
        # 1. a lambda function taking a group as argument and returning a dict of name: param instances
        # 2. a dict of subgroup name: subgroup_make functions
        # 3. a possibly-None parent group. If None, we are at root.
        root_group = RootParameterGroup(name, cplex_version, mdl)
        root_group._initialize(param_dict_fn, subgroup_fn)
        return root_group

    def prettyprint(self, indent=0):
        print("* CPLEX parameters version: {0}".format(self.cplex_version))
        ParameterGroup.prettyprint(self, indent)